from __future__ import annotations

from typing import Generator

from qlink_interface.interface import (
    ReqCreateBase,
    ResCreateAndKeep,
    ResMeasureDirectly,
)

from pydynaa import EventExpression
from qoala.runtime.environment import LocalEnvironment
from qoala.sim.common import ComponentProtocol, PortListener
from qoala.sim.egpmgr import EgpManager
from qoala.sim.memmgr import MemoryManager
from qoala.sim.message import Message
from qoala.sim.netstackcomp import NetstackComponent
from qoala.sim.qdevice import QDevice
from qoala.sim.signals import (
    SIGNAL_MEMORY_FREED,
    SIGNAL_PEER_NSTK_MSG,
    SIGNAL_PROC_NSTK_MSG,
)


class NetstackInterface(ComponentProtocol):
    """NetSquid protocol representing the QNodeOS network stack."""

    def __init__(
        self,
        comp: NetstackComponent,
        local_env: LocalEnvironment,
        qdevice: QDevice,
        memmgr: MemoryManager,
        egpmgr: EgpManager,
    ) -> None:
        """Network stack protocol constructor. Typically created indirectly through
        constructing a `Qnos` instance.

        :param comp: NetSquid component representing the network stack
        :param qnos: `Qnos` protocol that owns this protocol
        """
        super().__init__(name=f"{comp.name}_protocol", comp=comp)
        self._comp = comp
        self._qdevice = qdevice
        self._local_env = local_env
        self._memmgr = memmgr
        self._egpmgr = egpmgr

        self.add_listener(
            "qnos",
            PortListener(self._comp.qnos_in_port, SIGNAL_PROC_NSTK_MSG),
        )

        self.add_listener(
            "qnos_mem",
            PortListener(self._comp.qnos_mem_in_port, SIGNAL_MEMORY_FREED),
        )

        for peer in self._local_env.get_all_other_node_names():
            self.add_listener(
                f"peer_{peer}",
                PortListener(
                    self._comp.peer_in_port(peer), f"{SIGNAL_PEER_NSTK_MSG}_{peer}"
                ),
            )

    def send_qnos_msg(self, msg: Message) -> None:
        """Send a message to the processor."""
        self._comp.qnos_out_port.tx_output(msg)

    def receive_qnos_msg(self) -> Generator[EventExpression, None, Message]:
        """Receive a message from the processor. Block until there is at least one
        message."""
        return (yield from self._receive_msg("qnos", SIGNAL_PROC_NSTK_MSG))

    def send_peer_msg(self, peer: str, msg: Message) -> None:
        """Send a message to the network stack of the other node.

        NOTE: for now we assume there is only one other node, which is 'the' peer."""
        self._comp.peer_out_port(peer).tx_output(msg)

    def receive_peer_msg(self, peer: str) -> Generator[EventExpression, None, Message]:
        """Receive a message from the network stack of the other node. Block until
        there is at least one message.

        NOTE: for now we assume there is only one other node, which is 'the' peer."""
        return (
            yield from self._receive_msg(
                f"peer_{peer}", f"{SIGNAL_PEER_NSTK_MSG}_{peer}"
            )
        )

    def put_request(self, remote_id: int, request: ReqCreateBase) -> None:
        egp = self._egpmgr.get_egp(remote_id)
        egp.put(request)

    def await_result_create_keep(
        self, remote_id: int
    ) -> Generator[EventExpression, None, ResCreateAndKeep]:
        egp = self._egpmgr.get_egp(remote_id)
        yield self.await_signal(
            sender=egp,
            signal_label=ResCreateAndKeep.__name__,
        )
        result: ResCreateAndKeep = egp.get_signal_result(
            ResCreateAndKeep.__name__, receiver=self
        )
        return result

    def await_result_measure_directly(
        self, remote_id: int
    ) -> Generator[EventExpression, None, ResMeasureDirectly]:
        egp = self._egpmgr.get_egp(remote_id)
        yield self.await_signal(
            sender=egp,
            signal_label=ResMeasureDirectly.__name__,
        )
        result: ResMeasureDirectly = egp.get_signal_result(
            ResMeasureDirectly.__name__, receiver=self
        )
        return result

    def await_memory_freed_signal(
        self, pid: int, virt_id: int
    ) -> Generator[EventExpression, None, None]:
        # TODO: use pid and virt_id?
        yield from self._receive_msg("qnos_mem", SIGNAL_MEMORY_FREED)

    @property
    def qdevice(self) -> QDevice:
        return self._qdevice

    @property
    def memmgr(self) -> MemoryManager:
        return self._memmgr

    @property
    def egpmgr(self) -> EgpManager:
        return self._egpmgr

    def remote_id_to_peer_name(self, remote_id: int) -> str:
        node_info = self._local_env.get_global_env().get_nodes()[remote_id]
        # TODO figure out why mypy does not like this
        return node_info.name  # type: ignore



from netqasm.lang.instr.core import MeasInstruction, RetRegInstruction, SetInstruction
from netqasm.lang.instr.vanilla import RotZInstruction
from netqasm.lang.operand import Register, Template
from netqasm.lang.subroutine import Subroutine

from qoala.lang.iqoala import (
    AddCValueOp,
    AssignCValueOp,
    IqoalaProgram,
    IqoalaSharedMemLoc,
    IqoalaSubroutine,
    IqoalaVector,
    ProgramMeta,
    ReceiveCMsgOp,
    ReturnResultOp,
    RunSubroutineOp,
    SendCMsgOp,
)
from qoala.util.tests import text_equal


def test_serialize_meta_1():
    expected = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END
    """

    meta = ProgramMeta(name="alice", parameters=[], csockets={0: "bob"}, epr_sockets={})
    assert text_equal(meta.serialize(), expected)


def test_serialize_meta_2():
    expected = """
META_START
name: alice
parameters: theta1, theta2
csockets: 0 -> bob, 1 -> charlie
epr_sockets: 1 -> charlie
META_END
    """

    meta = ProgramMeta(
        name="alice",
        parameters=["theta1", "theta2"],
        csockets={0: "bob", 1: "charlie"},
        epr_sockets={1: "charlie"},
    )
    assert text_equal(meta.serialize(), expected)


def test_serialize_instructions_1():
    expected = """
my_value = assign_cval() : 1
remote_id = assign_cval() : 0
send_cmsg(remote_id, my_value)
received_value = recv_cmsg(remote_id)
new_value = assign_cval() : 3
my_value = add_cval_c(new_value, new_value)
vec<m> = run_subroutine(vec<my_value>) : subrt1
return_result(m)
    """

    instructions = [
        AssignCValueOp("my_value", 1),
        AssignCValueOp("remote_id", 0),
        SendCMsgOp("remote_id", "my_value"),
        ReceiveCMsgOp("remote_id", "received_value"),
        AssignCValueOp("new_value", 3),
        AddCValueOp("my_value", "new_value", "new_value"),
        RunSubroutineOp(IqoalaVector(["m"]), IqoalaVector(["my_value"]), "subrt1"),
        ReturnResultOp("m"),
    ]
    program = IqoalaProgram(
        instructions=instructions, subroutines={}, meta=ProgramMeta.empty("alice")
    )
    assert text_equal(program.serialize_instructions(), expected)


def test_serialize_subroutines_1():
    expected = """
SUBROUTINE subrt1
    params: my_value
    returns: M0 -> m
  NETQASM_START
    set Q0 0
    rot_z Q0 {my_value} 4
    meas Q0 M0
    ret_reg M0
  NETQASM_END
    """

    Q0 = Register.from_str("Q0")
    M0 = Register.from_str("M0")
    subrt = IqoalaSubroutine(
        name="subrt1",
        subrt=Subroutine(
            instructions=[
                SetInstruction(reg=Q0, imm=0),
                RotZInstruction(reg=Q0, imm0=Template("my_value"), imm1=4),
                MeasInstruction(reg0=Q0, reg1=M0),
                RetRegInstruction(reg=M0),
            ],
            arguments=["my_value"],
        ),
        return_map={"m": IqoalaSharedMemLoc("M0")},
    )
    program = IqoalaProgram(
        instructions=[], subroutines={"subrt1": subrt}, meta=ProgramMeta.empty("alice")
    )
    assert text_equal(program.serialize_subroutines(), expected)


def test_serialize_subroutines_2():
    expected = """
SUBROUTINE subrt1
    params: param1
    returns: M0 -> m
  NETQASM_START
    set R0 {param1}
    meas Q0 M0
  NETQASM_END

SUBROUTINE subrt2
    params: theta
    returns: 
  NETQASM_START
    set R0 {theta}
  NETQASM_END
    """

    R0 = Register.from_str("R0")
    Q0 = Register.from_str("Q0")
    M0 = Register.from_str("M0")
    subrt1 = IqoalaSubroutine(
        name="subrt1",
        subrt=Subroutine(
            instructions=[
                SetInstruction(reg=R0, imm=Template("param1")),
                MeasInstruction(reg0=Q0, reg1=M0),
            ],
            arguments=["param1"],
        ),
        return_map={"m": IqoalaSharedMemLoc("M0")},
    )
    subrt2 = IqoalaSubroutine(
        name="subrt2",
        subrt=Subroutine(
            instructions=[
                SetInstruction(reg=R0, imm=Template("theta")),
            ],
            arguments=["theta"],
        ),
        return_map={},
    )
    program = IqoalaProgram(
        instructions=[],
        subroutines={"subrt1": subrt1, "subrt2": subrt2},
        meta=ProgramMeta.empty("alice"),
    )
    assert text_equal(program.serialize_subroutines(), expected)


def test_serialize_program():
    expected = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END

my_value = assign_cval() : 1
remote_id = assign_cval() : 0
send_cmsg(remote_id, my_value)
received_value = recv_cmsg(remote_id)
new_value = assign_cval() : 3
my_value = add_cval_c(new_value, new_value)
vec<m> = run_subroutine(vec<my_value>) : subrt1
return_result(m)

SUBROUTINE subrt1
    params: my_value
    returns: M0 -> m
  NETQASM_START
    set Q0 0
    rot_z Q0 {my_value} 4
    meas Q0 M0
    ret_reg M0
  NETQASM_END
    """

    meta = ProgramMeta(name="alice", parameters=[], csockets={0: "bob"}, epr_sockets={})
    instructions = [
        AssignCValueOp("my_value", 1),
        AssignCValueOp("remote_id", 0),
        SendCMsgOp("remote_id", "my_value"),
        ReceiveCMsgOp("remote_id", "received_value"),
        AssignCValueOp("new_value", 3),
        AddCValueOp("my_value", "new_value", "new_value"),
        RunSubroutineOp(IqoalaVector(["m"]), IqoalaVector(["my_value"]), "subrt1"),
        ReturnResultOp("m"),
    ]
    Q0 = Register.from_str("Q0")
    M0 = Register.from_str("M0")
    subrt = IqoalaSubroutine(
        name="subrt1",
        subrt=Subroutine(
            instructions=[
                SetInstruction(reg=Q0, imm=0),
                RotZInstruction(reg=Q0, imm0=Template("my_value"), imm1=4),
                MeasInstruction(reg0=Q0, reg1=M0),
                RetRegInstruction(reg=M0),
            ],
            arguments=["my_value"],
        ),
        return_map={"m": IqoalaSharedMemLoc("M0")},
    )

    program = IqoalaProgram(
        instructions=instructions, subroutines={"subrt1": subrt}, meta=meta
    )

    assert text_equal(program.serialize(), expected)


if __name__ == "__main__":
    test_serialize_meta_1()
    test_serialize_meta_2()
    test_serialize_instructions_1()
    test_serialize_subroutines_1()
    test_serialize_subroutines_2()
    test_serialize_program()



SIGNAL_HOST_HOST_MSG = "EvHostHandMsg"
SIGNAL_HOST_HAND_MSG = "EvHostHandMsg"
SIGNAL_HAND_HOST_MSG = "EvHandHostMsg"
SIGNAL_HAND_PROC_MSG = "EvHandProcMsg"
SIGNAL_PROC_HAND_MSG = "EvProcHandMsg"
SIGNAL_PROC_NSTK_MSG = "EvProcNstkMsg"
SIGNAL_NSTK_PROC_MSG = "EvNstkProcMsg"
SIGNAL_PEER_NSTK_MSG = "EvPeerNstkMsg"

SIGNAL_MEMORY_FREED = "EvMemoryFreed"



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Dict, List, Tuple

import netsquid as ns
from netsquid.nodes import Node
from netsquid.qubits import ketstates, qubitapi
from netsquid_magic.link_layer import (
    MagicLinkLayerProtocolWithSignaling,
    SingleClickTranslationUnit,
)
from netsquid_magic.magic_distributor import PerfectStateMagicDistributor

from qoala.lang.iqoala import IqoalaParser, IqoalaProgram
from qoala.runtime.config import (
    GenericQDeviceConfig,
    LinkConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
)
from qoala.runtime.environment import GlobalEnvironment, GlobalNodeInfo
from qoala.runtime.program import BatchInfo, BatchResult, ProgramInput
from qoala.runtime.schedule import (
    ProgramTaskList,
    SchedulerInput,
    SchedulerOutput,
    SchedulerOutputEntry,
    ScheduleSolver,
    TaskBuilder,
)
from qoala.sim.build import build_network
from qoala.sim.egp import EgpProtocol


def create_global_env(
    num_qubits: int, names: List[str] = ["alice", "bob", "charlie"]
) -> GlobalEnvironment:

    env = GlobalEnvironment()
    for i, name in enumerate(names):
        env.add_node(i, GlobalNodeInfo(name, i))
    env.set_global_schedule([0, 1, 2])
    env.set_timeslot_len(1e6)
    return env


def create_egp_protocols(node1: Node, node2: Node) -> Tuple[EgpProtocol, EgpProtocol]:
    link_dist = PerfectStateMagicDistributor(nodes=[node1, node2], state_delay=0)
    link_prot = MagicLinkLayerProtocolWithSignaling(
        nodes=[node1, node2],
        magic_distributor=link_dist,
        translation_unit=SingleClickTranslationUnit(),
    )
    return EgpProtocol(node1, link_prot), EgpProtocol(node2, link_prot)


def create_server_tasks(
    server_program: IqoalaProgram, cfg: ProcNodeConfig
) -> ProgramTaskList:
    tasks = []

    cl_dur = 1e3
    cc_dur = 10e6
    # ql_dur = 1e4
    qc_dur = 1e6

    qdevice_cfg: GenericQDeviceConfig = cfg.qdevice_cfg

    set_dur = cfg.instr_latency
    rot_dur = qdevice_cfg.single_qubit_gate_time
    h_dur = qdevice_cfg.single_qubit_gate_time
    meas_dur = qdevice_cfg.measure_time
    free_dur = cfg.instr_latency
    cphase_dur = qdevice_cfg.two_qubit_gate_time

    # csocket = assign_cval() : 0
    tasks.append(TaskBuilder.CL(cl_dur, 0))
    # run_subroutine(vec<client_id>) : create_epr_0
    tasks.append(TaskBuilder.CL(cl_dur, 1))
    tasks.append(TaskBuilder.QC(qc_dur, "create_epr_0"))
    # run_subroutine(vec<client_id>) : create_epr_1
    tasks.append(TaskBuilder.CL(cl_dur, 2))
    tasks.append(TaskBuilder.QC(qc_dur, "create_epr_1"))
    # run_subroutine(vec<client_id>) : local_cphase
    tasks.append(TaskBuilder.CL(cl_dur, 3))
    tasks.append(TaskBuilder.QL(set_dur, "local_cphase", 0))
    tasks.append(TaskBuilder.QL(set_dur, "local_cphase", 1))
    tasks.append(TaskBuilder.QL(cphase_dur, "local_cphase", 2))
    # delta1 = recv_cmsg(client_id)
    tasks.append(TaskBuilder.CC(cc_dur, 4))
    # vec<m1> = run_subroutine(vec<delta1>) : meas_qubit_1
    tasks.append(TaskBuilder.CL(cl_dur, 5))
    tasks.append(TaskBuilder.QL(set_dur, "meas_qubit_1", 0))
    tasks.append(TaskBuilder.QL(rot_dur, "meas_qubit_1", 1))
    tasks.append(TaskBuilder.QL(h_dur, "meas_qubit_1", 2))
    tasks.append(TaskBuilder.QL(meas_dur, "meas_qubit_1", 3))
    tasks.append(TaskBuilder.QL(free_dur, "meas_qubit_1", 4))
    # send_cmsg(csocket, m1)
    tasks.append(TaskBuilder.CC(cc_dur, 6))
    # delta2 = recv_cmsg(csocket)
    tasks.append(TaskBuilder.CC(cc_dur, 7))
    # vec<m2> = run_subroutine(vec<delta2>) : meas_qubit_0
    tasks.append(TaskBuilder.CL(cl_dur, 8))
    tasks.append(TaskBuilder.QL(set_dur, "meas_qubit_0", 0))
    tasks.append(TaskBuilder.QL(rot_dur, "meas_qubit_0", 1))
    tasks.append(TaskBuilder.QL(h_dur, "meas_qubit_0", 2))
    tasks.append(TaskBuilder.QL(meas_dur, "meas_qubit_0", 3))
    tasks.append(TaskBuilder.QL(free_dur, "meas_qubit_0", 4))
    # return_result(m1)
    tasks.append(TaskBuilder.CL(cl_dur, 9))
    # return_result(m2)
    tasks.append(TaskBuilder.CL(cl_dur, 10))

    return ProgramTaskList(server_program, {i: task for i, task in enumerate(tasks)})


def create_client_tasks(
    client_program: IqoalaProgram, cfg: ProcNodeConfig
) -> ProgramTaskList:
    tasks = []

    cl_dur = 1e3
    cc_dur = 10e6
    # ql_dur = 1e3
    qc_dur = 1e6

    qdevice_cfg: GenericQDeviceConfig = cfg.qdevice_cfg

    set_dur = cfg.instr_latency
    rot_dur = qdevice_cfg.single_qubit_gate_time
    h_dur = qdevice_cfg.single_qubit_gate_time
    meas_dur = qdevice_cfg.measure_time
    free_dur = cfg.instr_latency

    tasks.append(TaskBuilder.CL(cl_dur, 0))
    tasks.append(TaskBuilder.CL(cl_dur, 1))
    tasks.append(TaskBuilder.QC(qc_dur, "create_epr_0"))
    tasks.append(TaskBuilder.CL(cl_dur, 2))
    tasks.append(TaskBuilder.QL(set_dur, "post_epr_0", 0))
    tasks.append(TaskBuilder.QL(rot_dur, "post_epr_0", 1))
    tasks.append(TaskBuilder.QL(h_dur, "post_epr_0", 2))
    tasks.append(TaskBuilder.QL(meas_dur, "post_epr_0", 3))
    tasks.append(TaskBuilder.QL(free_dur, "post_epr_0", 4))

    tasks.append(TaskBuilder.CL(cl_dur, 3))
    tasks.append(TaskBuilder.QC(qc_dur, "create_epr_1"))
    tasks.append(TaskBuilder.CL(cl_dur, 4))
    tasks.append(TaskBuilder.QL(set_dur, "post_epr_1", 0))
    tasks.append(TaskBuilder.QL(rot_dur, "post_epr_1", 1))
    tasks.append(TaskBuilder.QL(h_dur, "post_epr_1", 2))
    tasks.append(TaskBuilder.QL(meas_dur, "post_epr_1", 3))
    tasks.append(TaskBuilder.QL(free_dur, "post_epr_1", 4))

    tasks.append(TaskBuilder.CL(cl_dur, 5))
    tasks.append(TaskBuilder.CL(cl_dur, 6))
    tasks.append(TaskBuilder.CL(cl_dur, 7))
    tasks.append(TaskBuilder.CL(cl_dur, 8))
    tasks.append(TaskBuilder.CC(cc_dur, 9))
    tasks.append(TaskBuilder.CC(cc_dur, 10))
    tasks.append(TaskBuilder.CL(cl_dur, 11))
    tasks.append(TaskBuilder.CL(cl_dur, 12))
    tasks.append(TaskBuilder.CL(cl_dur, 13))
    tasks.append(TaskBuilder.CL(cl_dur, 14))
    tasks.append(TaskBuilder.CL(cl_dur, 15))
    tasks.append(TaskBuilder.CC(cc_dur, 16))
    tasks.append(TaskBuilder.CL(cl_dur, 17))
    tasks.append(TaskBuilder.CL(cl_dur, 18))

    return ProgramTaskList(client_program, {i: task for i, task in enumerate(tasks)})


class NaiveSolver(ScheduleSolver):
    @classmethod
    def solve(cls, input: SchedulerInput) -> SchedulerOutput:
        output_entries: List[SchedulerOutputEntry] = []

        assert len(input.num_executions) == input.num_programs
        assert len(input.num_instructions) == input.num_programs
        assert len(input.instr_durations) == input.num_programs

        current_time = 0

        for i in range(input.num_programs):
            num_executions = input.num_executions[i]
            num_instructions = input.num_instructions[i]
            instr_durations = input.instr_durations[i]
            for j in range(num_executions):
                for k in range(num_instructions):
                    duration = instr_durations[k]
                    entry = SchedulerOutputEntry(
                        app_index=i,
                        ex_index=j,
                        instr_index=k,
                        start_time=current_time,
                    )
                    current_time += duration
                    output_entries.append(entry)

        return SchedulerOutput(output_entries)


class NoTimeSolver(ScheduleSolver):
    @classmethod
    def solve(cls, input: SchedulerInput) -> SchedulerOutput:
        output_entries: List[SchedulerOutputEntry] = []

        assert len(input.num_executions) == input.num_programs
        assert len(input.num_instructions) == input.num_programs
        assert len(input.instr_durations) == input.num_programs

        current_time = 0

        for i in range(input.num_programs):
            num_executions = input.num_executions[i]
            num_instructions = input.num_instructions[i]
            instr_durations = input.instr_durations[i]
            for j in range(num_executions):
                for k in range(num_instructions):
                    duration = instr_durations[k]
                    entry = SchedulerOutputEntry(
                        app_index=i,
                        ex_index=j,
                        instr_index=k,
                        start_time=None,
                    )
                    current_time += duration
                    output_entries.append(entry)

        return SchedulerOutput(output_entries)


@dataclass
class BqcResult:
    client_results: Dict[int, BatchResult]
    server_results: Dict[int, BatchResult]


def run_bqc(alpha, beta, theta1, theta2, num_iterations: int):
    num_qubits = 3
    global_env = create_global_env(num_qubits, names=["client", "server"])
    server_id = global_env.get_node_id("server")
    client_id = global_env.get_node_id("client")

    server_node_cfg = ProcNodeConfig(
        name="server",
        node_id=server_id,
        qdevice_typ="generic",
        qdevice_cfg=GenericQDeviceConfig.perfect_config(num_qubits),
        instr_latency=1000,
    )
    client_node_cfg = ProcNodeConfig(
        name="client",
        node_id=client_id,
        qdevice_typ="generic",
        qdevice_cfg=GenericQDeviceConfig.perfect_config(num_qubits),
        instr_latency=1000,
    )
    link_cfg = LinkConfig.perfect_config("server", "client")

    network_cfg = ProcNodeNetworkConfig(
        nodes=[server_node_cfg, client_node_cfg], links=[link_cfg]
    )
    network = build_network(network_cfg, global_env)
    server_procnode = network.nodes["server"]
    client_procnode = network.nodes["client"]

    path = os.path.join(os.path.dirname(__file__), "test_bqc_server.iqoala")
    with open(path) as file:
        server_text = file.read()
    server_program = IqoalaParser(server_text).parse()
    server_tasks = create_server_tasks(server_program, server_node_cfg)
    server_inputs = [
        ProgramInput({"client_id": client_id}) for _ in range(num_iterations)
    ]
    server_batch_info = BatchInfo(
        program=server_program,
        inputs=server_inputs,
        num_iterations=num_iterations,
        deadline=0,
        tasks=server_tasks,
        num_qubits=3,
    )
    server_procnode.submit_batch(server_batch_info)
    server_procnode.initialize_processes()
    server_procnode.initialize_schedule(NoTimeSolver)

    path = os.path.join(os.path.dirname(__file__), "test_bqc_client.iqoala")
    with open(path) as file:
        client_text = file.read()
    client_program = IqoalaParser(client_text).parse()
    client_tasks = create_client_tasks(client_program, client_node_cfg)
    client_inputs = [
        ProgramInput(
            {
                "server_id": server_id,
                "alpha": alpha,
                "beta": beta,
                "theta1": theta1,
                "theta2": theta2,
            }
        )
        for _ in range(num_iterations)
    ]

    client_batch_info = BatchInfo(
        program=client_program,
        inputs=client_inputs,
        num_iterations=num_iterations,
        deadline=0,
        tasks=client_tasks,
        num_qubits=3,
    )
    client_procnode.submit_batch(client_batch_info)
    client_procnode.initialize_processes()
    client_procnode.initialize_schedule(NoTimeSolver)

    server_procnode.start()
    client_procnode.start()
    ns.sim_run()

    client_results = client_procnode.scheduler.get_batch_results()
    server_results = server_procnode.scheduler.get_batch_results()

    return BqcResult(client_results, server_results)


def expected_rsp_qubit(theta: int, p: int, dummy: bool):
    expected = qubitapi.create_qubits(1)[0]

    if dummy:
        if p == 0:
            qubitapi.assign_qstate(expected, ketstates.s0)
        elif p == 1:
            qubitapi.assign_qstate(expected, ketstates.s1)
    else:
        if (theta, p) == (0, 0):
            qubitapi.assign_qstate(expected, ketstates.h0)
        elif (theta, p) == (0, 1):
            qubitapi.assign_qstate(expected, ketstates.h1)
        if (theta, p) == (8, 0):
            qubitapi.assign_qstate(expected, ketstates.y0)
        elif (theta, p) == (8, 1):
            qubitapi.assign_qstate(expected, ketstates.y1)
        if (theta, p) == (16, 0):
            qubitapi.assign_qstate(expected, ketstates.h1)
        elif (theta, p) == (16, 1):
            qubitapi.assign_qstate(expected, ketstates.h0)
        if (theta, p) == (-8, 0):
            qubitapi.assign_qstate(expected, ketstates.y1)
        elif (theta, p) == (-8, 1):
            qubitapi.assign_qstate(expected, ketstates.y0)

    return expected


def expected_rsp_state(theta: int, p: int, dummy: bool):
    expected = qubitapi.create_qubits(1)[0]

    if dummy:
        if p == 0:
            return ketstates.s0
        elif p == 1:
            return ketstates.s1
    else:
        if (theta, p) == (0, 0):
            return ketstates.h0
        elif (theta, p) == (0, 1):
            return ketstates.h1
        if (theta, p) == (8, 0):
            return ketstates.y0
        elif (theta, p) == (8, 1):
            return ketstates.y1
        if (theta, p) == (16, 0):
            return ketstates.h1
        elif (theta, p) == (16, 1):
            return ketstates.h0
        if (theta, p) == (-8, 0):
            return ketstates.y1
        elif (theta, p) == (-8, 1):
            return ketstates.y0

    return expected.qstate


def test_bqc():

    # Effective computation: measure in Z the following state:
    # H Rz(beta) H Rz(alpha) |+>
    # m2 should be this outcome

    # angles are in multiples of pi/16

    # LogManager.set_log_level("DEBUG")
    # LogManager.log_to_file("test_run.log")

    def check(alpha, beta, theta1, theta2, expected):
        ns.sim_reset()
        bqc_result = run_bqc(
            alpha=alpha, beta=beta, theta1=theta1, theta2=theta2, num_iterations=20
        )

        server_batch_results = bqc_result.server_results
        for batch_id, batch_results in server_batch_results.items():
            program_results = batch_results.results
            m2s = [result.values["m2"] for result in program_results]
            assert all(m2 == expected for m2 in m2s)

    check(alpha=8, beta=8, theta1=0, theta2=0, expected=0)
    check(alpha=8, beta=24, theta1=0, theta2=0, expected=1)
    check(alpha=8, beta=8, theta1=13, theta2=27, expected=0)
    check(alpha=8, beta=24, theta1=2, theta2=22, expected=1)


if __name__ == "__main__":
    test_bqc()



from dataclasses import dataclass
from enum import Enum, auto
from typing import Generator, List, Optional, Set

from netsquid.components.instructions import INSTR_INIT, Instruction
from netsquid.components.qprocessor import QuantumProcessor
from netsquid.components.qprogram import QuantumProgram
from netsquid.nodes import Node
from netsquid.qubits.qubit import Qubit

from pydynaa import EventExpression


class QDeviceType(Enum):
    GENERIC = 0
    NV = auto()


class PhysicalQuantumMemory:
    def __init__(self, comm_ids: Set[int], mem_ids: Set[int]) -> None:
        self._comm_qubit_ids: Set[int] = comm_ids
        self._mem_qubit_ids: Set[int] = mem_ids
        self._all_ids = comm_ids.union(mem_ids)
        self._qubit_count = len(self._all_ids)

    @property
    def qubit_count(self) -> int:
        return self._qubit_count

    @property
    def comm_qubit_count(self) -> int:
        return len(self._comm_qubit_ids)

    @property
    def comm_qubit_ids(self) -> Set[int]:
        return self._comm_qubit_ids

    @property
    def mem_qubit_ids(self) -> Set[int]:
        return self._mem_qubit_ids


class GenericPhysicalQuantumMemory(PhysicalQuantumMemory):
    def __init__(self, qubit_count: int) -> None:
        super().__init__(
            comm_ids={i for i in range(qubit_count)},
            mem_ids={i for i in range(qubit_count)},
        )


class NVPhysicalQuantumMemory(PhysicalQuantumMemory):
    def __init__(self, qubit_count: int) -> None:
        super().__init__(comm_ids={0}, mem_ids={i for i in range(1, qubit_count)})


class UnsupportedQDeviceCommandError(Exception):
    pass


class NonInitializedQubitError(Exception):
    pass


@dataclass(eq=True, frozen=True)
class QDeviceCommand:
    instr: Instruction
    indices: List[int]
    angle: Optional[float] = None


class QDevice:
    def __init__(
        self, node: Node, typ: QDeviceType, memory: PhysicalQuantumMemory
    ) -> None:
        self._node = node
        self._typ = typ
        self._memory = memory

    @property
    def qprocessor(self) -> QuantumProcessor:
        """Get the NetSquid `QuantumProcessor` object of this node."""
        return self._node.qmemory

    @property
    def typ(self) -> QDeviceType:
        return self._typ

    @property
    def memory(self) -> PhysicalQuantumMemory:
        return self._memory

    @property
    def qubit_count(self) -> int:
        return self.memory.qubit_count

    @property
    def comm_qubit_count(self) -> int:
        return self.memory.comm_qubit_count

    @property
    def comm_qubit_ids(self) -> Set[int]:
        return self.memory.comm_qubit_ids

    @property
    def mem_qubit_ids(self) -> Set[int]:
        return self.memory.mem_qubit_ids

    @property
    def all_qubit_ids(self) -> Set[int]:
        return self.comm_qubit_ids.union(self.mem_qubit_ids)

    def is_allowed(self, cmd: QDeviceCommand) -> bool:
        all_phys_instructions = self.qprocessor.get_physical_instructions()

        # Get the physical instruction with the same type ('gate').
        matches = [
            i for i in all_phys_instructions if i.instruction.name == cmd.instr.name
        ]
        # Should be at least one matching.
        if len(matches) == 0:
            return False

        for phys_instr in matches:
            # If there is no topology, this instruction is allowed on any qubit.
            if phys_instr.topology is None:
                return True

            # Else, check if it is allowed for our current qubit(s).
            if len(cmd.indices) == 1:
                if cmd.indices[0] in phys_instr.topology:
                    return True
            else:
                if (cmd.indices[0], cmd.indices[1]) in phys_instr.topology:
                    return True

        # We didn't find any matching instruction.
        return False

    def set_mem_pos_in_use(self, id: int, in_use: bool) -> None:
        self.qprocessor.mem_positions[id].in_use = in_use

    def execute_commands(
        self, commands: List[QDeviceCommand]
    ) -> Generator[EventExpression, None, Optional[int]]:
        """Can only return at most 1 measurement result."""
        prog = QuantumProgram()

        # TODO: rewrite this abomination

        for cmd in commands:
            # Check if this instruction is allowed on this processor.
            # If not, NetSquid will just silently skip this instruction which is confusing.
            if not self.is_allowed(cmd):
                raise UnsupportedQDeviceCommandError

        for cmd in commands:
            # Check if the qubit has been initialized, since instructions won't work
            # if this is not the case.
            # Anything after an INSTR_INIT instruction is fine.
            # TODO: better logic for detecting INITs of individual qubits.
            if cmd.instr == INSTR_INIT:
                break
            for index in cmd.indices:
                if self.get_local_qubit(index) is None:
                    raise NonInitializedQubitError

        for cmd in commands:
            if cmd.angle is not None:
                prog.apply(cmd.instr, qubit_indices=cmd.indices, angle=cmd.angle)
            else:
                prog.apply(cmd.instr, qubit_indices=cmd.indices)
        yield self.qprocessor.execute_program(prog)

        last_result = prog.output["last"]
        if last_result is not None:
            meas_outcome: int = last_result[0]
            return meas_outcome
        return None

    def execute_program(
        self, prog: QuantumProgram
    ) -> Generator[EventExpression, None, None]:
        raise DeprecationWarning

    def get_local_qubit(self, index: int) -> Qubit:
        return self.qprocessor.peek([index], skip_noise=True)[0]

    def get_local_qubits(self, indices: List[int]) -> List[Qubit]:
        return self.qprocessor.peek(indices, skip_noise=True)  # type: ignore



from __future__ import annotations

from typing import Dict, Optional, Type

from netsquid.components import QuantumProcessor
from netsquid.protocols import Protocol
from netsquid_magic.link_layer import MagicLinkLayerProtocolWithSignaling

from qoala.runtime.environment import GlobalEnvironment, LocalEnvironment
from qoala.runtime.program import BatchInfo, ProgramBatch
from qoala.runtime.schedule import Schedule, ScheduleSolver
from qoala.sim.egp import EgpProtocol
from qoala.sim.egpmgr import EgpManager
from qoala.sim.host import Host
from qoala.sim.hostcomp import HostComponent
from qoala.sim.memmgr import MemoryManager
from qoala.sim.netstack import Netstack, NetstackComponent
from qoala.sim.process import IqoalaProcess
from qoala.sim.procnodecomp import ProcNodeComponent
from qoala.sim.qdevice import (
    GenericPhysicalQuantumMemory,
    NVPhysicalQuantumMemory,
    QDevice,
    QDeviceType,
)
from qoala.sim.qnos import Qnos
from qoala.sim.qnoscomp import QnosComponent
from qoala.sim.scheduler import Scheduler


class ProcNode(Protocol):
    """NetSquid protocol representing a node with a software stack."""

    def __init__(
        self,
        name: str,
        global_env: GlobalEnvironment,
        qprocessor: QuantumProcessor,
        node: Optional[ProcNodeComponent] = None,
        qdevice_type: Optional[str] = "generic",
        node_id: Optional[int] = None,
        scheduler: Optional[Scheduler] = None,
        asynchronous: bool = False,
    ) -> None:
        """ProcNode constructor.

        :param name: name of this node
        :param node: an existing ProcNodeComponent object containing the static
            components or None. If None, a ProcNodeComponent is automatically
            created.
        :param qdevice_type: hardware type of the QDevice, defaults to "generic"
        :param qprocessor: NetSquid `QuantumProcessor` representing the QDevice,
            defaults to None. If None, a QuantumProcessor is created
            automatically.
        :param node_id: ID to use for the internal NetSquid node object
        :param use_default_components: whether to automatically create NetSquid
            components for the Host and QNodeOS, defaults to True. If False,
            this allows for manually creating and adding these components.
        """
        super().__init__(name=f"{name}")
        if node:
            self._node = node
        else:
            self._node = ProcNodeComponent(name, qprocessor, global_env, node_id)

        self._global_env = global_env
        self._local_env = LocalEnvironment(global_env, global_env.get_node_id(name))
        self._asynchronous = asynchronous

        # Create internal components.
        self._qdevice: QDevice
        if qdevice_type == "generic":
            physical_memory = GenericPhysicalQuantumMemory(qprocessor.num_positions)
            self._qdevice = QDevice(self._node, QDeviceType.GENERIC, physical_memory)
        elif qdevice_type == "nv":
            physical_memory = NVPhysicalQuantumMemory(qprocessor.num_positions)
            self._qdevice = QDevice(self._node, QDeviceType.NV, physical_memory)
        else:
            raise ValueError

        self._host = Host(self.host_comp, self._local_env, self._asynchronous)
        self._memmgr = MemoryManager(self.node.name, self._qdevice)
        self._egpmgr = EgpManager()
        self._qnos = Qnos(
            self.qnos_comp,
            self._local_env,
            self._memmgr,
            self._qdevice,
            self._asynchronous,
        )
        self._netstack = Netstack(
            self.netstack_comp,
            self._local_env,
            self._memmgr,
            self._egpmgr,
            self._qdevice,
        )

        if scheduler is None:
            self._scheduler = Scheduler(
                self._node.name,
                self._host,
                self._qnos,
                self._netstack,
                self._memmgr,
                self._local_env,
            )
        else:
            self._scheduler = scheduler

    def install_schedule(self, schedule: Schedule) -> None:
        self.scheduler.install_schedule(schedule)

    def assign_ll_protocol(
        self, remote_id: int, prot: MagicLinkLayerProtocolWithSignaling
    ) -> None:
        """Set the link layer protocol to use for entanglement generation.

        The same link layer protocol object is used by both nodes sharing a link in
        the network."""
        self.egpmgr.add_egp(remote_id, EgpProtocol(self.node, prot))

    @property
    def node(self) -> ProcNodeComponent:
        return self._node

    @property
    def host_comp(self) -> HostComponent:
        return self.node.host_comp

    @property
    def qnos_comp(self) -> QnosComponent:
        return self.node.qnos_comp

    @property
    def netstack_comp(self) -> NetstackComponent:
        return self.node.netstack_comp

    @property
    def qdevice(self) -> QDevice:
        return self._qdevice

    @qdevice.setter
    def qdevice(self, qdevice) -> QDevice:
        self._qdevice = qdevice
        self.qnos.qdevice = qdevice
        self.netstack.qdevice = qdevice

    @property
    def host(self) -> Host:
        return self._host

    @host.setter
    def host(self, host: Host) -> None:
        self._host = host

    @property
    def qnos(self) -> Qnos:
        return self._qnos

    @qnos.setter
    def qnos(self, qnos: Qnos) -> None:
        self._qnos = qnos

    @property
    def netstack(self) -> Netstack:
        return self._netstack

    @netstack.setter
    def netstack(self, netstack: Netstack) -> None:
        self._netstack = netstack

    @property
    def memmgr(self) -> MemoryManager:
        return self._memmgr

    @property
    def egpmgr(self) -> EgpManager:
        return self._egpmgr

    @property
    def scheduler(self) -> Scheduler:
        return self._scheduler

    @scheduler.setter
    def scheduler(self, scheduler: Scheduler) -> None:
        self._scheduler = scheduler

    def connect_to(self, other: ProcNode) -> None:
        """Create connections between ports of this ProcNode and those of
        another ProcNode."""
        here = self.node.name
        there = other.node.name
        self.node.host_peer_out_port(there).connect(other.node.host_peer_in_port(here))
        self.node.host_peer_in_port(there).connect(other.node.host_peer_out_port(here))
        self.node.netstack_peer_out_port(there).connect(
            other.node.netstack_peer_in_port(here)
        )
        self.node.netstack_peer_in_port(there).connect(
            other.node.netstack_peer_out_port(here)
        )

    def start(self) -> None:
        assert self._host is not None
        assert self._qnos is not None
        assert self._netstack is not None
        super().start()
        self._host.start()
        self._qnos.start()
        self._netstack.start()
        self._scheduler.start()

    def stop(self) -> None:
        assert self._host is not None
        assert self._qnos is not None
        assert self._netstack is not None
        self._scheduler.stop()
        self._netstack.stop()
        self._qnos.stop()
        self._host.stop()
        super().stop()

    def submit_batch(self, batch_info: BatchInfo) -> None:
        self.scheduler.submit_batch(batch_info)

    def initialize_processes(self) -> None:
        self.scheduler.create_processes_for_batches()

    def initialize_schedule(self, solver: Type[ScheduleSolver]) -> None:
        self.scheduler.solve_and_install_schedule(solver)

    def add_process(self, process: IqoalaProcess) -> None:
        self.memmgr.add_process(process)

    def get_batches(self) -> Dict[int, ProgramBatch]:
        return self.scheduler.get_batches()



import logging
from typing import Generator, List, Optional

import netsquid as ns
from netqasm.sdk.build_epr import (
    SER_RESPONSE_KEEP_IDX_BELL_STATE,
    SER_RESPONSE_KEEP_IDX_GOODNESS,
    SER_RESPONSE_KEEP_LEN,
    SER_RESPONSE_MEASURE_IDX_MEASUREMENT_BASIS,
    SER_RESPONSE_MEASURE_IDX_MEASUREMENT_OUTCOME,
    SER_RESPONSE_MEASURE_LEN,
)
from netsquid.components.instructions import INSTR_ROT_X, INSTR_ROT_Z
from netsquid.qubits.ketstates import BellIndex
from qlink_interface import (
    ReqCreateAndKeep,
    ReqCreateBase,
    ReqMeasureDirectly,
    ReqReceive,
    ReqRemoteStatePrep,
    ResCreateAndKeep,
    ResMeasureDirectly,
)

from pydynaa import EventExpression
from qoala.sim.constants import PI
from qoala.sim.logging import LogManager
from qoala.sim.memmgr import AllocError
from qoala.sim.memory import ProgramMemory, SharedMemory
from qoala.sim.message import Message
from qoala.sim.netstackinterface import NetstackInterface
from qoala.sim.process import IqoalaProcess
from qoala.sim.qdevice import QDevice, QDeviceCommand
from qoala.sim.requests import (
    EprCreateType,
    NetstackBreakpointCreateRequest,
    NetstackBreakpointReceiveRequest,
    NetstackCreateRequest,
    NetstackReceiveRequest,
    T_NetstackRequest,
)


class NetstackProcessor:
    def __init__(self, interface: NetstackInterface) -> None:
        self._interface = interface

        self._name = f"{interface.name}_NetstackProcessor"

        self._logger: logging.Logger = LogManager.get_stack_logger(  # type: ignore
            f"{self.__class__.__name__}({self._name})"
        )

        # memory of current program, only not-None when processor is active
        self._current_prog_mem: Optional[ProgramMemory] = None

    def _prog_mem(self) -> ProgramMemory:
        # May only be called when processor is active
        assert self._current_prog_mem is not None
        return self._current_prog_mem

    @property
    def qdevice(self) -> QDevice:
        return self._interface.qdevice

    def get_shared_mem(self, pid: int) -> SharedMemory:
        prog_mem = self._interface.memmgr.get_process(pid).prog_memory
        return prog_mem.shared_mem

    def _create_link_layer_create_request(
        # self, remote_id: int, args: List[int], epr_socket: EprSocket
        self,
        request: NetstackCreateRequest,
    ) -> ReqCreateBase:
        """Construct a link layer request from application request info.

        :param remote_id: ID of remote node
        :param args: NetQASM array elements from the arguments array specified by the
            application
        :return: link layer request object
        """

        if request.typ == EprCreateType.CREATE_KEEP:
            ll_request = ReqCreateAndKeep(
                remote_node_id=request.remote_id,
                number=request.num_pairs,
                minimum_fidelity=request.fidelity,
            )
        elif request.typ == EprCreateType.MEASURE_DIRECTLY:
            ll_request = ReqMeasureDirectly(
                remote_node_id=request.remote_id,
                number=request.num_pairs,
                minimum_fidelity=request.fidelity,
            )
        elif request.typ == EprCreateType.REMOTE_STATE_PREP:
            ll_request = ReqRemoteStatePrep(
                remote_node_id=request.remote_id,
                number=request.num_pairs,
                minimum_fidelity=request.fidelity,
            )
        else:
            raise ValueError(f"Unsupported create type {request.typ}")
        return ll_request

    def assign(
        self, process: IqoalaProcess, request: T_NetstackRequest
    ) -> Generator[EventExpression, None, None]:

        if isinstance(request, NetstackCreateRequest):
            yield from self.handle_create_request(process, request)
            self._logger.debug("create request done")
        elif isinstance(request, NetstackReceiveRequest):
            yield from self.handle_receive_request(process, request)
            self._logger.debug("receive request done")
        elif isinstance(request, NetstackBreakpointCreateRequest):
            yield from self.handle_breakpoint_create_request(request)
            self._logger.debug("breakpoint create request done")
        elif isinstance(request, NetstackBreakpointReceiveRequest):
            yield from self.handle_breakpoint_receive_request(request)
            self._logger.debug("breakpoint receive request done")
        else:
            raise RuntimeError

    def handle_create_request(
        self, process: IqoalaProcess, req: NetstackCreateRequest
    ) -> Generator[EventExpression, None, None]:
        """Issue a request to create entanglement with a remote node.

        :param req: request info
        """

        # Synchronize with the remote node.

        # Send the request to the receiver node and wait for an acknowledgement.
        peer = self._interface.remote_id_to_peer_name(req.remote_id)
        self._interface.send_peer_msg(peer, Message(content=req))
        peer_msg = yield from self._interface.receive_peer_msg(peer)
        self._logger.debug(f"received peer msg: {peer_msg}")

        # Handle the request.
        if req.typ == EprCreateType.CREATE_KEEP:
            yield from self.handle_create_ck_request(process, req)
        elif req.typ == EprCreateType.MEASURE_DIRECTLY:
            yield from self.handle_create_md_request(process, req)
        else:
            raise RuntimeError

    def handle_create_ck_request(
        self, process: IqoalaProcess, req: NetstackCreateRequest
    ) -> Generator[EventExpression, None, None]:
        """Handle a Create and Keep request as the initiator/creator, until all
        pairs have been created.

        This method uses the EGP protocol to create and measure EPR pairs with
        the remote node. It will fully complete the request before returning. If
        the pair created by the EGP protocol is another Bell state than Phi+,
        local gates are applied to do a correction, such that the final
        delivered pair is always Phi+.

        The method can however yield (i.e. give control back to the simulator
        scheduler) in the following cases: - no communication qubit is
        available; this method will resume when a
          SIGNAL_MEMORY_FREED is given (currently only the processor can do
          this)
        - when waiting for the EGP protocol to produce the next pair; this
          method resumes when the pair is delivered
        - a Bell correction gate is applied

        This method does not return anything. This method has the side effect
        that NetQASM array value are written to.

        :param req: application request info (app ID and NetQASM array IDs)
        :param request: link layer request object
        """
        num_pairs = req.num_pairs

        self._logger.info(f"putting CK request to EGP for {num_pairs} pairs")
        self._logger.info(f"qubit IDs specified by application: {req.virt_qubit_ids}")
        self._logger.info(f"splitting request into {num_pairs} 1-pair requests")

        start_time = ns.sim_time()

        for pair_index in range(num_pairs):
            virt_id = req.virt_qubit_ids[pair_index]
            ll_result = yield from self.create_single_pair(
                process, req, virt_id, wait_for_free=True
            )

            gen_duration = ns.sim_time() - start_time

            self.write_pair_result(
                process,
                ll_result,
                pair_index,
                req.result_array_addr,
                gen_duration,
            )

            self._interface.send_qnos_msg(Message(content="wrote to array"))

    def create_single_pair(
        self,
        process: IqoalaProcess,
        request: NetstackCreateRequest,
        virt_id: int,
        wait_for_free: bool = False,
    ) -> Generator[EventExpression, None, ResCreateAndKeep]:
        """
        :param wait_for_free: whether to wait (block) on a "memory free" signal in case
            of an AllocError
        """
        ll_request = self._create_link_layer_create_request(request)
        ll_request.number = 1

        self._logger.info("trying to allocate comm qubit")
        while True:
            try:
                self._interface.memmgr.allocate_comm(process.pid, virt_id)
                break
            except AllocError:
                if not wait_for_free:
                    raise AllocError

                # else:
                self._logger.info("no comm qubit available, waiting...")

                # Wait for a signal indicating the communication qubit might be free
                # again.
                yield from self._interface.await_memory_freed_signal(
                    process.pid, virt_id
                )
                self._logger.info(
                    "a 'free' happened, trying again to allocate comm qubit..."
                )

        # Put the request to the EGP.
        self._logger.info("putting CK request")
        self._interface.put_request(request.remote_id, ll_request)

        # Wait for a signal from the EGP.
        self._logger.info("waiting for result")
        result = yield from self._interface.await_result_create_keep(request.remote_id)
        self._logger.info(f"got result: {result}")

        # Bell state corrections. Resulting state is always Phi+ (i.e. B00).
        if result.bell_state == BellIndex.B00:
            pass
        elif result.bell_state == BellIndex.B01:
            commands = [QDeviceCommand(INSTR_ROT_X, [0], angle=PI)]
            yield from self._interface.qdevice.execute_commands(commands)
        elif result.bell_state == BellIndex.B10:
            commands = [QDeviceCommand(INSTR_ROT_Z, [0], angle=PI)]
            yield from self._interface.qdevice.execute_commands(commands)
        elif result.bell_state == BellIndex.B11:
            commands = [
                QDeviceCommand(INSTR_ROT_X, [0], angle=PI),
                QDeviceCommand(INSTR_ROT_Z, [0], angle=PI),
            ]
            yield from self._interface.qdevice.execute_commands(commands)

        return result

    def write_pair_result(
        self,
        process: IqoalaProcess,
        ll_result: ResCreateAndKeep,
        pair_index: int,
        array_addr: int,
        duration: float,  # in ns
    ) -> None:
        shared_mem = process.prog_memory.shared_mem

        gen_duration_us_int = int(duration / 1000)
        self._logger.info(f"gen duration (us): {gen_duration_us_int}")

        # Length of response array slice for a single pair.
        slice_len = SER_RESPONSE_KEEP_LEN

        # Populate results array.
        for i in range(slice_len):
            # Write -1 to unused array elements.
            value = -1

            # Write corresponding result value to the other array elements.
            if i == SER_RESPONSE_KEEP_IDX_GOODNESS:
                value = gen_duration_us_int
            if i == SER_RESPONSE_KEEP_IDX_BELL_STATE:
                value = ll_result.bell_state

            # Calculate array element location.
            arr_index = slice_len * pair_index + i

            shared_mem.set_array_value(array_addr, arr_index, value)
        self._logger.debug(
            f"wrote to @{array_addr}[{slice_len * pair_index}:"
            f"{slice_len * pair_index + slice_len}] for app ID {process.pid}"
        )

    def handle_create_md_request(
        self, req: NetstackCreateRequest, request: ReqMeasureDirectly
    ) -> Generator[EventExpression, None, None]:
        """Handle a Create and Measure request as the initiator/creator, until all
        pairs have been created and measured.

        This method uses the EGP protocol to create EPR pairs with the remote node.
        It will fully complete the request before returning.

        No Bell state corrections are done. This means that application code should
        use the result information to check, for each pair, the generated Bell state
        and possibly post-process the measurement outcomes.

        The method can yield (i.e. give control back to the simulator scheduler) in
        the following cases:
        - no communication qubit is available; this method will resume when a
          SIGNAL_MEMORY_FREED is given (currently only the processor can do this)
        - when waiting for the EGP protocol to produce the next pair; this method
          resumes when the pair is delivered

        This method does not return anything.
        This method has the side effect that NetQASM array value are written to.

        :param req: application request info (app ID and NetQASM array IDs)
        :param request: link layer request object
        """

        # Put the reqeust to the EGP.
        self._interface.put_request(req.remote_node_id, request)

        results: List[ResMeasureDirectly] = []

        # Wait for all pairs to be created. For each pair, the EGP sends a separate
        # signal that is awaited here. Only after the last pair, we write the results
        # to the array. This is done since the whole request (i.e. all pairs) is
        # expected to finish in a short time anyway. However, writing results for a
        # pair as soon as they are done may be implemented in the future.
        for _ in range(request.number):
            self._interface.memmgr.allocate(req.pid, 0)

            result = yield from self._interface.await_result_measure_directly(
                req.remote_node_id
            )
            self._logger.debug(f"bell index: {result.bell_state}")
            results.append(result)
            self._interface.memmgr.free(req.pid, 0)

        shared_mem = self.get_shared_mem(req.pid)

        # Length of response array slice for a single pair.
        slice_len = SER_RESPONSE_MEASURE_LEN

        # Populate results array.
        for pair_index in range(request.number):
            result = results[pair_index]

            for i in range(slice_len):
                # Write -1 to unused array elements.
                value = -1

                # Write corresponding result value to the other array elements.
                if i == SER_RESPONSE_MEASURE_IDX_MEASUREMENT_OUTCOME:
                    value = result.measurement_outcome
                elif i == SER_RESPONSE_MEASURE_IDX_MEASUREMENT_BASIS:
                    value = result.measurement_basis.value
                elif i == SER_RESPONSE_KEEP_IDX_BELL_STATE:
                    value = result.bell_state.value

                # Calculate array element location.
                arr_index = slice_len * pair_index + i

                shared_mem.set_array_value(req.result_array_addr, arr_index, value)

        self._interface.send_qnos_msg(Message(content="wrote to array"))

    def handle_receive_ck_request(
        self, process: IqoalaProcess, req: NetstackReceiveRequest
    ) -> Generator[EventExpression, None, None]:
        """Handle a Create and Keep request as the receiver, until all pairs have
        been created.

        This method uses the EGP protocol to create EPR pairs with the remote
        node. It will fully complete the request before returning.

        If the pair created by the EGP protocol is another Bell state than Phi+,
        it is assumed that the *other* node applies local gates such that the
        final delivered pair is always Phi+.

        The method can yield (i.e. give control back to the simulator scheduler)
        in the following cases: - no communication qubit is available; this
        method will resume when a
          SIGNAL_MEMORY_FREED is given (currently only the processor can do
          this)
        - when waiting for the EGP protocol to produce the next pair; this
          method resumes when the pair is delivered

        This method does not return anything. This method has the side effect
        that NetQASM array value are written to.

        :param req: application request info (app ID and NetQASM array IDs)
        :param request: link layer request object
        """

        num_pairs = req.num_pairs

        self._logger.info(f"putting CK request to EGP for {num_pairs} pairs")
        self._logger.info(f"splitting request into {num_pairs} 1-pair requests")

        start_time = ns.sim_time()

        for pair_index in range(num_pairs):
            virt_id = req.virt_qubit_ids[pair_index]
            ll_result = yield from self.receive_single_pair(
                process, req, virt_id, wait_for_free=True
            )

            gen_duration = ns.sim_time() - start_time
            self.write_pair_result(
                process, ll_result, pair_index, req.result_array_addr, gen_duration
            )

            self._interface.send_qnos_msg(Message(content="wrote to array"))

    def receive_single_pair(
        self,
        process: IqoalaProcess,
        request: NetstackReceiveRequest,
        virt_id: int,
        wait_for_free: bool = False,
    ) -> Generator[EventExpression, None, ResCreateAndKeep]:
        self._logger.info("trying to allocate comm qubit")
        while True:
            try:
                self._interface.memmgr.allocate_comm(process.pid, virt_id)
                break
            except AllocError:
                if not wait_for_free:
                    raise AllocError

                self._logger.info("no comm qubit available, waiting...")

                # Wait for a signal indicating the communication qubit might be free
                # again.
                yield from self._interface.await_memory_freed_signal(
                    process.pid, virt_id
                )
                self._logger.info(
                    "a 'free' happened, trying again to allocate comm qubit..."
                )

        # Put the request to the EGP.
        self._logger.info("putting CK request")
        self._interface.put_request(request.remote_id, ReqReceive(request.remote_id))
        self._logger.info("waiting for result")

        # Wait for a signal from the EGP.
        result = yield from self._interface.await_result_create_keep(request.remote_id)
        self._logger.info(f"got result: {result}")

        return result

    def handle_receive_md_request(
        self, req: NetstackReceiveRequest, request: ReqMeasureDirectly
    ) -> Generator[EventExpression, None, None]:
        """Handle a Create and Measure request as the receiver, until all
        pairs have been created and measured.

        This method uses the EGP protocol to create EPR pairs with the remote node.
        It will fully complete the request before returning.

        No Bell state corrections are done. This means that application code should
        use the result information to check, for each pair, the generated Bell state
        and possibly post-process the measurement outcomes.

        The method can yield (i.e. give control back to the simulator scheduler)
        in the following cases: - no communication qubit is available; this
        method will resume when a
          SIGNAL_MEMORY_FREED is given (currently only the processor can do
          this)
        - when waiting for the EGP protocol to produce the next pair; this
          method resumes when the pair is delivered

        This method does not return anything. This method has the side effect
        that NetQASM array value are written to.

        :param req: application request info (app ID and NetQASM array IDs)
        :param request: link layer request object
        """
        assert isinstance(request, ReqMeasureDirectly)

        self._interface.put_request(req.remote_node_id, ReqReceive(req.remote_node_id))

        results: List[ResMeasureDirectly] = []

        for _ in range(request.number):
            self._interface.memmgr.allocate(req.pid, 0)

            result = yield from self._interface.await_result_measure_directly(
                req.remote_node_id
            )
            results.append(result)

            self._interface.memmgr.free(req.pid, 0)

        shared_mem = self.get_shared_mem(req.pid)

        # Length of response array slice for a single pair.
        slice_len = SER_RESPONSE_MEASURE_LEN

        # Populate results array.
        for pair_index in range(request.number):
            result = results[pair_index]

            for i in range(slice_len):
                # Write -1 to unused array elements.
                value = -1

                # Write corresponding result value to the other array elements.
                if i == SER_RESPONSE_MEASURE_IDX_MEASUREMENT_OUTCOME:
                    value = result.measurement_outcome
                elif i == SER_RESPONSE_MEASURE_IDX_MEASUREMENT_BASIS:
                    value = result.measurement_basis.value
                elif i == SER_RESPONSE_KEEP_IDX_BELL_STATE:
                    value = result.bell_state.value

                # Calculate array element location.
                arr_index = slice_len * pair_index + i

                shared_mem.set_array_value(req.result_array_addr, arr_index, value)

            self._interface.send_qnos_msg(Message(content="wrote to array"))

    def handle_receive_request(
        self, process: IqoalaProcess, req: NetstackReceiveRequest
    ) -> Generator[EventExpression, None, None]:
        """Issue a request to receive entanglement from a remote node.

        :param req: request info
        """

        # Synchronize with the remote node.

        # Wait for the network stack in the remote node to get the corresponding
        # 'create' request from its local application and send it to us.
        # NOTE: we do not check if the request from the other node matches our own
        # request. Also, we simply block until synchronizing with the other node,
        # and then fully handle the request. There is no support for queueing
        # and/or interleaving multiple different requests.
        peer = self._interface.remote_id_to_peer_name(req.remote_id)
        msg = yield from self._interface.receive_peer_msg(peer)
        create_request = msg.content
        self._logger.debug(f"received {create_request} from peer")

        # Acknowledge to the remote node that we received the request and we will
        # start handling it.
        self._logger.debug("sending 'ready' to peer")
        self._interface.send_peer_msg(peer, Message(content="ready"))

        if req.typ == EprCreateType.CREATE_KEEP:
            yield from self.handle_receive_ck_request(process, req)
        elif req.typ == EprCreateType.MEASURE_DIRECTLY:
            yield from self.handle_receive_md_request(process, req)
        else:
            raise RuntimeError

    def handle_breakpoint_create_request(
        self, request: NetstackBreakpointCreateRequest
    ) -> Generator[EventExpression, None, None]:
        # Use epr sockets for this process to get all relevant remote nodes.
        epr_sockets = self._interface.memmgr.get_process(request.pid).epr_sockets
        remote_ids = [esck.remote_id for esck in epr_sockets.values()]
        remote_names = [self._interface.remote_id_to_peer_name(id) for id in remote_ids]

        # Synchronize with the remote nodes.
        for peer in remote_names:
            self._interface.send_peer_msg(peer, Message(content="breakpoint start"))

        for peer in remote_names:
            response = yield from self._interface.receive_peer_msg(peer)
            assert response.content == "breakpoint start"

        # Remote nodes are now ready. Notify the processor.
        self._interface.send_qnos_msg(Message(content="breakpoint ready"))

        # Wait for the processor to finish handling the breakpoint.
        processor_msg = yield from self._interface.receive_qnos_msg()
        assert processor_msg.content == "breakpoint end"

        # Tell the remote nodes that the breakpoint has finished.
        for peer in remote_names:
            self._interface.send_peer_msg(peer, Message(content="breakpoint end"))

        # Wait for the remote node to have finsihed as well.
        for peer in remote_names:
            response = yield from self._interface.receive_peer_msg(peer)
            assert response.content == "breakpoint end"

        # Notify the processor that we are done.
        self._interface.send_qnos_msg(Message(content="breakpoint finished"))

    def handle_breakpoint_receive_request(
        self, request: NetstackBreakpointReceiveRequest
    ) -> Generator[EventExpression, None, None]:
        # Use epr sockets for this process to get all relevant remote nodes.
        epr_sockets = self._interface.memmgr.get_process(request.pid).epr_sockets
        remote_ids = [esck.remote_id for esck in epr_sockets.values()]
        remote_names = [self._interface.remote_id_to_peer_name(id) for id in remote_ids]

        # Synchronize with the remote nodes.
        for peer in remote_names:
            msg = yield from self._interface.receive_peer_msg(peer)
            assert msg.content == "breakpoint start"

        for peer in remote_names:
            self._interface.send_peer_msg(peer, Message(content="breakpoint start"))

        # Notify the processor we are ready to handle the breakpoint.
        self._interface.send_qnos_msg(Message(content="breakpoint ready"))

        # Wait for the processor to finish handling the breakpoint.
        processor_msg = yield from self._interface.receive_qnos_msg()
        assert processor_msg.content == "breakpoint end"

        # Wait for the remote nodes to finish and tell it we are finished as well.
        for peer in remote_names:
            peer_msg = yield from self._interface.receive_peer_msg(peer)
            assert peer_msg.content == "breakpoint end"

        for peer in remote_names:
            self._interface.send_peer_msg(peer, Message(content="breakpoint end"))

        # Notify the processor that we are done.
        self._interface.send_qnos_msg(Message(content="breakpoint finished"))



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from qoala.runtime.environment import GlobalEnvironment
from qoala.sim.globals import GlobalSimData


@dataclass
class SimulationContext:
    global_env: Optional[GlobalEnvironment] = None
    global_sim_data: Optional[GlobalSimData] = None



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto
from typing import Dict, List, Optional, Tuple, Union

from qoala.lang.iqoala import IqoalaProgram


class ProcessorType(Enum):
    HOST = 0
    QNOS = auto()
    NETSTACK = auto()


class InstructionType(Enum):
    CL = 0
    CC = auto()
    QL = auto()
    QC = auto()


@dataclass
class HostTask:
    instr_index: int


@dataclass
class QnosTask:
    subrt_name: str
    instr_index: int


@dataclass
class NetstackTask:
    # request_name: str  # TODO: needed?
    subrt_name: str


@dataclass
class SingleProgramTask:
    instr_type: InstructionType
    processor_type: ProcessorType
    instr_index: Optional[int]
    subrt_name: Optional[str]
    request_name: Optional[str]
    duration: int

    def as_host_task(self) -> HostTask:
        assert self.processor_type == ProcessorType.HOST
        assert self.instr_index is not None
        return HostTask(self.instr_index)

    def as_qnos_task(self) -> QnosTask:
        assert self.processor_type == ProcessorType.QNOS
        assert self.subrt_name is not None
        assert self.instr_index is not None
        return QnosTask(self.subrt_name, self.instr_index)

    def as_netstack_task(self) -> NetstackTask:
        assert self.processor_type == ProcessorType.NETSTACK
        assert self.subrt_name is not None
        return NetstackTask(self.subrt_name)

    def __str__(self) -> str:
        return f"{self.processor_type.name} {self.subrt_name} {self.instr_index}"


@dataclass
class CombinedProgramTask:
    tasks: List[SingleProgramTask]

    @property
    def instr_type(self) -> InstructionType:
        assert len(self.tasks) > 0
        return self.tasks[0].instr_type

    @property
    def duration(self) -> int:
        assert len(self.tasks) > 0
        return sum(task.duration for task in self.tasks)


ProgramTask = Union[SingleProgramTask, CombinedProgramTask]


class TaskBuilder:
    @classmethod
    def CC(cls, duration, index: int) -> ProgramTask:
        return SingleProgramTask(
            instr_type=InstructionType.CC,
            processor_type=ProcessorType.HOST,
            instr_index=index,
            subrt_name=None,
            request_name=None,
            duration=duration,
        )

    @classmethod
    def CL(cls, duration, index: int) -> ProgramTask:
        return SingleProgramTask(
            instr_type=InstructionType.CL,
            processor_type=ProcessorType.HOST,
            instr_index=index,
            subrt_name=None,
            request_name=None,
            duration=duration,
        )

    @classmethod
    def QL(cls, duration, subrt_name: str, index: int) -> ProgramTask:
        return SingleProgramTask(
            instr_type=InstructionType.QL,
            processor_type=ProcessorType.QNOS,
            instr_index=index,
            subrt_name=subrt_name,
            request_name=None,
            duration=duration,
        )

    @classmethod
    def QC(cls, duration, subrt_name: str) -> ProgramTask:
        return SingleProgramTask(
            instr_type=InstructionType.QC,
            processor_type=ProcessorType.NETSTACK,
            instr_index=None,
            subrt_name=subrt_name,
            request_name=None,
            duration=duration,
        )

    @classmethod
    def task_group(cls, tasks: List[SingleProgramTask]) -> CombinedProgramTask:
        assert len(tasks) > 0
        typ = tasks[0].instr_type
        assert all(t.instr_type == typ for t in tasks)
        return CombinedProgramTask(tasks)


@dataclass
class ProgramTaskList:
    program: IqoalaProgram
    tasks: Dict[int, ProgramTask]  # task index -> task

    @classmethod
    def empty(cls, program: IqoalaProgram) -> ProgramTaskList:
        return ProgramTaskList(program, {})


@dataclass
class SchedulerInput:
    global_schedule: List[int]
    timeslot_len: int
    num_programs: int
    deadlines: List[int]
    num_executions: List[int]
    num_instructions: List[int]
    instr_durations: List[List[int]]
    instr_types: List[List[InstructionType]]


@dataclass
class SchedulerOutputEntry:
    app_index: int
    ex_index: int
    instr_index: int
    start_time: Optional[int]  # None means "earliest time possible"


@dataclass
class SchedulerOutput:
    entries: List[SchedulerOutputEntry]


@dataclass(eq=True, frozen=True)
class ScheduleEntry:
    pid: int
    task_index: int


@dataclass(eq=True, frozen=True)
class ScheduleTime:
    time: Optional[int]  # None means "earliest time possible"


@dataclass
class Schedule:
    entries: List[Tuple[ScheduleTime, ScheduleEntry]]  # list of (time, entry)


class ScheduleSolver:
    @classmethod
    def solve(cls, input: SchedulerInput) -> SchedulerOutput:
        raise NotImplementedError


class NaiveSolver(ScheduleSolver):
    """
    Schedules all tasks as a consecutive list without gaps.
    No interleaving of programs of program iterations.
    No adherence to any global schedule.
    """

    @classmethod
    def solve(cls, input: SchedulerInput) -> SchedulerOutput:
        output_entries: List[SchedulerOutputEntry] = []

        assert len(input.num_executions) == input.num_programs
        assert len(input.num_instructions) == input.num_programs
        assert len(input.instr_durations) == input.num_programs

        current_time = 0

        for i in range(input.num_programs):
            num_executions = input.num_executions[i]
            num_instructions = input.num_instructions[i]
            instr_durations = input.instr_durations[i]
            for j in range(num_executions):
                for k in range(num_instructions):
                    duration = instr_durations[k]
                    entry = SchedulerOutputEntry(
                        app_index=i,
                        ex_index=j,
                        instr_index=k,
                        start_time=current_time,
                    )
                    current_time += duration
                    output_entries.append(entry)

        return SchedulerOutput(output_entries)


class NoTimeSolver(ScheduleSolver):
    """
    No time constraints at all.
    Schedules all tasks as a consecutive list without gaps.
    No interleaving of programs of program iterations.
    No adherence to any global schedule.
    """

    @classmethod
    def solve(cls, input: SchedulerInput) -> SchedulerOutput:
        output_entries: List[SchedulerOutputEntry] = []

        assert len(input.num_executions) == input.num_programs
        assert len(input.num_instructions) == input.num_programs
        assert len(input.instr_durations) == input.num_programs

        current_time = 0

        for i in range(input.num_programs):
            num_executions = input.num_executions[i]
            num_instructions = input.num_instructions[i]
            instr_durations = input.instr_durations[i]
            for j in range(num_executions):
                for k in range(num_instructions):
                    duration = instr_durations[k]
                    entry = SchedulerOutputEntry(
                        app_index=i,
                        ex_index=j,
                        instr_index=k,
                        start_time=None,
                    )
                    current_time += duration
                    output_entries.append(entry)

        return SchedulerOutput(output_entries)



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto
from typing import Dict, List, Optional, Tuple, Union

from netqasm.lang.instr import NetQASMInstruction
from netqasm.lang.instr.flavour import NVFlavour
from netqasm.lang.operand import Template
from netqasm.lang.parsing.text import parse_text_subroutine
from netqasm.lang.subroutine import Subroutine

from qoala.sim.requests import (
    EprCreateRole,
    EprCreateType,
    NetstackCreateRequest,
    NetstackReceiveRequest,
    T_NetstackRequest,
)


@dataclass
class ProgramMeta:
    name: str
    parameters: List[str]  # list of parameter names (all have type int)
    csockets: Dict[int, str]  # socket ID -> remote node name
    epr_sockets: Dict[int, str]  # socket ID -> remote node name

    @classmethod
    def empty(cls, name: str) -> ProgramMeta:
        return ProgramMeta(name=name, parameters=[], csockets={}, epr_sockets={})

    def serialize(self) -> str:
        s = "META_START"
        s += f"\nname: {self.name}"
        s += f"\nparameters: {', '.join(self.parameters)}"
        s += f"\ncsockets: {', '.join(f'{k} -> {v}' for k,v in self.csockets.items())}"
        s += f"\nepr_sockets: {', '.join(f'{k} -> {v}' for k,v in self.epr_sockets.items())}"
        s += "\nMETA_END"
        return s


class IqoalaInstructionType(Enum):
    CC = 0
    CL = auto()
    QC = auto()
    QL = auto()


@dataclass
class IqoalaInstructionSignature:
    typ: IqoalaInstructionType
    duration: int = 0


class StaticIqoalaProgramInfo:
    pass


class DynamicIqoalaProgramInfo:
    pass


class IqoalaSubroutine:
    def __init__(
        self,
        name: str,
        subrt: Subroutine,
        return_map: Dict[str, IqoalaSharedMemLoc],
        request_name: Optional[str] = None,
    ) -> None:
        self._name = name
        self._subrt = subrt
        self._return_map = return_map
        self._request_name = request_name

    @property
    def name(self) -> str:
        return self._name

    @property
    def subroutine(self) -> Subroutine:
        return self._subrt

    @property
    def return_map(self) -> Dict[str, IqoalaSharedMemLoc]:
        return self._return_map

    @property
    def request_name(self) -> Optional[str]:
        return self._request_name

    def serialize(self) -> str:
        s = f"SUBROUTINE {self.name}"
        s += f"\nparams: {', '.join(self.subroutine.arguments)}"
        rm = self.return_map  # just to make next line fit on one line
        s += f"\nreturns: {', '.join(f'{v} -> {k}' for k, v in rm.items())}"
        s += "\nNETQASM_START\n"
        s += self.subroutine.print_instructions()
        s += "\nNETQASM_END"
        return s

    def __str__(self) -> str:
        s = "\n"
        for key, value in self.return_map.items():
            s += f"return {str(value)} -> {key}\n"
        s += "NETQASM_START\n"
        s += self.subroutine.print_instructions()
        s += "\nNETQASM_END"
        return s

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, IqoalaSubroutine):
            return NotImplemented
        return (
            self.name == other.name
            and self.subroutine == other.subroutine
            and self.return_map == other.return_map
        )


class IqoalaRequest:
    def __init__(
        self, name: str, role: EprCreateRole, request: T_NetstackRequest
    ) -> None:
        self._name = name
        self._role = role
        self._request = request

    @property
    def name(self) -> str:
        return self._name

    @property
    def role(self) -> EprCreateRole:
        return self._role

    @property
    def request(self) -> T_NetstackRequest:
        return self._request

    def serialize(self) -> str:
        s = f"REQUEST {self.name}"
        s += f"role: {self.role}"
        s += f"remote_id: {self.request.remote_id}"
        s += f"epr_socket_id: {self.request.epr_socket_id}"
        s += f"typ: {self.request.typ.name}"
        s += f"num_pairs: {self.request.num_pairs}"
        s += f"fidelity: {self.request.fidelity}"
        s += f"virt_qubit_ids: {','.join(self.request.virt_qubit_ids)}"
        s += f"result_array_addr: {self.request.result_array_addr}"
        return s

    def __str__(self) -> str:
        return self.serialize()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, IqoalaRequest):
            return NotImplemented
        return (
            self.name == other.name
            and self.role == other.role
            and self.request == other.request
        )


IqoalaValue = Union[int, Template, str]


class IqoalaAttribute:
    def __init__(self, value: IqoalaValue) -> None:
        self._value = value

    @property
    def value(self) -> IqoalaValue:
        return self._value


@dataclass(eq=True, frozen=True)
class IqoalaSharedMemLoc:
    loc: str

    def __str__(self) -> str:
        return str(self.loc)


class IqoalaVector:
    def __init__(self, values: List[str]) -> None:
        self._values = values

    @property
    def values(self) -> List[str]:
        return self._values

    def __str__(self) -> str:
        return f"vec<{','.join(v for v in self.values)}>"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, IqoalaVector):
            return NotImplemented
        return self.values == other.values


class ClassicalIqoalaOp:
    OP_NAME: str = None  # type: ignore
    TYP: IqoalaInstructionType = None  # type: ignore

    def __init__(
        self,
        arguments: Optional[Union[List[str], List[IqoalaVector]]] = None,
        results: Optional[List[str]] = None,
        attributes: Optional[List[IqoalaValue]] = None,
    ) -> None:
        # TODO: support list of strs and vectors
        # currently not needed and confuses mypy
        self._arguments: Union[List[str], List[IqoalaVector]]
        self._results: List[str]
        self._attributes: List[IqoalaValue]

        if arguments is None:
            self._arguments = []  # type: ignore
        else:
            self._arguments = arguments

        if results is None:
            self._results = []
        else:
            self._results = results

        if attributes is None:
            self._attributes = []
        else:
            self._attributes = attributes

    def __str__(self) -> str:
        results = ", ".join(str(r) for r in self.results)
        args = ", ".join(str(a) for a in self.arguments)
        attrs = ", ".join(str(a) for a in self.attributes)
        s = ""
        if len(results) > 0:
            s += f"{results} = "

        s += f"{self.op_name}({args})"

        if len(attrs) > 0:
            s += f" : {attrs}"
        return s

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ClassicalIqoalaOp):
            return NotImplemented
        return (
            self.results == other.results
            and self.arguments == other.arguments
            and self.attributes == other.attributes
        )

    @classmethod
    def from_generic_args(
        cls, result: Optional[str], args: List[str], attr: Optional[IqoalaValue]
    ) -> ClassicalIqoalaOp:
        raise NotImplementedError

    @property
    def op_name(self) -> str:
        return self.__class__.OP_NAME  # type: ignore

    @property
    def arguments(self) -> Union[List[str], List[IqoalaVector]]:
        return self._arguments

    @property
    def results(self) -> List[str]:
        return self._results

    @property
    def attributes(self) -> List[IqoalaValue]:
        return self._attributes


class AssignCValueOp(ClassicalIqoalaOp):
    OP_NAME = "assign_cval"
    TYP = IqoalaInstructionType.CL

    def __init__(self, result: str, value: IqoalaValue) -> None:
        super().__init__(results=[result], attributes=[value])

    @classmethod
    def from_generic_args(
        cls, result: Optional[str], args: List[str], attr: Optional[IqoalaValue]
    ):
        assert result is not None
        assert len(args) == 0
        assert attr is not None
        return cls(result, attr)


class SendCMsgOp(ClassicalIqoalaOp):
    OP_NAME = "send_cmsg"
    TYP = IqoalaInstructionType.CC

    def __init__(self, csocket: str, value: str) -> None:
        # args:
        #   csocket (int): ID of csocket
        #   value (str): name of variable holding the value to send
        super().__init__(arguments=[csocket, value])

    @classmethod
    def from_generic_args(
        cls, result: Optional[str], args: List[str], attr: Optional[IqoalaValue]
    ):
        assert result is None
        assert len(args) == 2
        assert attr is None
        return cls(args[0], args[1])


class ReceiveCMsgOp(ClassicalIqoalaOp):
    OP_NAME = "recv_cmsg"
    TYP = IqoalaInstructionType.CC

    def __init__(self, csocket: str, result: str) -> None:
        super().__init__(arguments=[csocket], results=[result])

    @classmethod
    def from_generic_args(
        cls, result: Optional[str], args: List[str], attr: Optional[IqoalaValue]
    ):
        assert result is not None
        assert len(args) == 1
        assert attr is None
        return cls(args[0], result)


class AddCValueOp(ClassicalIqoalaOp):
    OP_NAME = "add_cval_c"
    TYP = IqoalaInstructionType.CL

    def __init__(self, result: str, value0: str, value1: str) -> None:
        super().__init__(arguments=[value0, value1], results=[result])

    @classmethod
    def from_generic_args(
        cls, result: Optional[str], args: List[str], attr: Optional[IqoalaValue]
    ):
        assert result is not None
        assert len(args) == 2
        assert attr is None
        return cls(result, args[0], args[1])


class MultiplyConstantCValueOp(ClassicalIqoalaOp):
    OP_NAME = "mult_const"
    TYP = IqoalaInstructionType.CL

    def __init__(self, result: str, value0: str, const: IqoalaValue) -> None:
        # result = value0 * const
        super().__init__(arguments=[value0], attributes=[const], results=[result])

    @classmethod
    def from_generic_args(
        cls, result: Optional[str], args: List[str], attr: Optional[IqoalaValue]
    ):
        assert result is not None
        assert len(args) == 1
        assert attr is not None
        return cls(result, args[0], attr)


class BitConditionalMultiplyConstantCValueOp(ClassicalIqoalaOp):
    OP_NAME = "bcond_mult_const"
    TYP = IqoalaInstructionType.CL

    def __init__(self, result: str, value0: str, cond: str, const: IqoalaValue) -> None:
        # if const == 1:
        #   result = value0 * const
        # else:
        #   result = value0
        super().__init__(arguments=[value0, cond], attributes=[const], results=[result])

    @classmethod
    def from_generic_args(
        cls, result: Optional[str], args: List[str], attr: Optional[IqoalaValue]
    ):
        assert result is not None
        assert len(args) == 2
        assert attr is not None
        return cls(result, args[0], args[1], attr)


class RunSubroutineOp(ClassicalIqoalaOp):
    OP_NAME = "run_subroutine"
    TYP = IqoalaInstructionType.CL

    def __init__(self, result: IqoalaVector, values: IqoalaVector, subrt: str) -> None:
        super().__init__(results=[result], arguments=[values], attributes=[subrt])

    @classmethod
    def from_generic_args(
        cls, result: Optional[str], args: List[str], attr: Optional[IqoalaValue]
    ):
        if result is not None:
            assert isinstance(result, IqoalaVector)
        assert len(args) == 1
        assert isinstance(args[0], IqoalaVector)
        assert isinstance(attr, str)
        return cls(result, args[0], attr)

    @property
    def subroutine(self) -> str:
        assert isinstance(self.attributes[0], str)
        return self.attributes[0]

    def __str__(self) -> str:
        return super().__str__()


class ReturnResultOp(ClassicalIqoalaOp):
    OP_NAME = "return_result"
    TYP = IqoalaInstructionType.CL

    def __init__(self, value: str) -> None:
        super().__init__(arguments=[value])

    @classmethod
    def from_generic_args(
        cls, result: Optional[str], args: List[str], attr: Optional[IqoalaValue]
    ):
        assert result is None
        assert len(args) == 1
        assert attr is None
        return cls(args[0])


LHR_OP_NAMES: Dict[str, ClassicalIqoalaOp] = {
    cls.OP_NAME: cls  # type: ignore
    for cls in [
        SendCMsgOp,
        ReceiveCMsgOp,
        AddCValueOp,
        MultiplyConstantCValueOp,
        BitConditionalMultiplyConstantCValueOp,
        AssignCValueOp,
        RunSubroutineOp,
        ReturnResultOp,
    ]
}


def netqasm_instr_to_type(instr: NetQASMInstruction) -> IqoalaInstructionType:
    if instr.mnemonic in ["create_epr", "recv_epr"]:
        return IqoalaInstructionType.QC
    else:
        return IqoalaInstructionType.QL


class IqoalaProgram:
    def __init__(
        self,
        instructions: List[ClassicalIqoalaOp],
        subroutines: Dict[str, IqoalaSubroutine],
        meta: ProgramMeta,
        requests: Optional[Dict[str, IqoalaRequest]] = None,
    ) -> None:
        self._instructions: List[ClassicalIqoalaOp] = instructions
        self._subroutines: Dict[str, IqoalaSubroutine] = subroutines
        self._meta: ProgramMeta = meta

        if requests is None:
            self._requests: Dict[str, IqoalaRequest] = {}
        else:
            self._requests: Dict[str, IqoalaRequest] = requests

    @property
    def meta(self) -> ProgramMeta:
        return self._meta

    @property
    def instructions(self) -> List[ClassicalIqoalaOp]:
        return self._instructions

    @instructions.setter
    def instructions(self, new_instrs) -> None:
        self._instructions = new_instrs

    def get_instr_signatures(self) -> List[IqoalaInstructionSignature]:
        sigs: List[IqoalaInstructionSignature] = []
        for instr in self.instructions:
            if isinstance(instr, RunSubroutineOp):
                subrt = instr.subroutine
                for nq_instr in subrt.subroutine.instructions:
                    typ = netqasm_instr_to_type(nq_instr)
                    # TODO: add duration
                    sigs.append(IqoalaInstructionSignature(typ))
            else:
                sigs.append(IqoalaInstructionSignature(instr.TYP))
        return sigs

    @property
    def subroutines(self) -> Dict[str, IqoalaSubroutine]:
        return self._subroutines

    @subroutines.setter
    def subroutines(self, new_subroutines: Dict[str, IqoalaSubroutine]) -> None:
        self._subroutines = new_subroutines

    @property
    def requests(self) -> Dict[str, IqoalaRequest]:
        return self._requests

    @requests.setter
    def requests(self, new_requests: Dict[str, IqoalaRequest]) -> None:
        self._requests = new_requests

    def __str__(self) -> str:
        # self.me
        # instrs = [
        #     f"{str(i)}\n{self.subroutines[i.arguments[0]]}"  # inline subroutine contents
        #     if isinstance(i, RunSubroutineOp)
        #     else str(i)
        #     for i in self.instructions
        # ]

        # return "\n".join("  " + i for i in instrs)
        return "\n".join("  " + str(i) for i in self.instructions)

    def serialize_meta(self) -> str:
        return self.meta.serialize()

    def serialize_instructions(self) -> str:
        return "\n".join("  " + str(i) for i in self.instructions)

    def serialize_subroutines(self) -> str:
        return "\n".join(s.serialize() for s in self.subroutines.values())

    def serialize(self) -> str:
        return (
            self.meta.serialize()
            + "\n"
            + self.serialize_instructions()
            + "\n"
            + self.serialize_subroutines()
        )


class EndOfTextException(Exception):
    pass


class IqoalaParseError(Exception):
    pass


class IqoalaMetaParser:
    def __init__(self, text: str) -> None:
        self._text = text
        lines = [line.strip() for line in text.split("\n")]
        self._lines = [line for line in lines if len(line) > 0]
        self._lineno: int = 0

    def _next_line(self) -> None:
        self._lineno += 1

    def _read_line(self) -> str:
        while True:
            if self._lineno >= len(self._lines):
                raise EndOfTextException
            line = self._lines[self._lineno]
            self._next_line()
            if len(line) > 0:
                return line
            # if no non-empty line, will always break on EndOfLineException

    def _parse_meta_line(self, key: str, line: str) -> List[str]:
        split = line.split(":")
        assert len(split) >= 1
        assert split[0] == key
        if len(split) == 1:
            return []
        assert len(split) == 2
        if len(split[1]) == 0:
            return []
        values = split[1].split(",")
        return [v.strip() for v in values]

    def _parse_meta_mapping(self, value_str: str) -> Dict[int, str]:
        result_dict = {}
        for v in value_str:
            key_value = [x.strip() for x in v.split("->")]
            assert len(key_value) == 2
            result_dict[int(key_value[0].strip())] = key_value[1].strip()
        return result_dict

    def parse(self) -> ProgramMeta:
        try:
            start_line = self._read_line()
            assert start_line == "META_START"

            name_values = self._parse_meta_line("name", self._read_line())
            assert len(name_values) == 1
            name = name_values[0]

            parameters = self._parse_meta_line("parameters", self._read_line())

            csockets_map = self._parse_meta_line("csockets", self._read_line())
            csockets = self._parse_meta_mapping(csockets_map)
            epr_sockets_map = self._parse_meta_line("epr_sockets", self._read_line())
            epr_sockets = self._parse_meta_mapping(epr_sockets_map)

            end_line = self._read_line()
            if end_line != "META_END":
                raise IqoalaParseError("Could not parse meta.")
        except AssertionError:
            raise IqoalaParseError
        except EndOfTextException:
            raise IqoalaParseError

        return ProgramMeta(name, parameters, csockets, epr_sockets)


class IqoalaInstrParser:
    def __init__(self, text: str) -> None:
        self._text = text
        lines = [line.strip() for line in text.split("\n")]
        self._lines = [line for line in lines if len(line) > 0]
        self._lineno: int = 0

    def _next_line(self) -> None:
        self._lineno += 1

    def _read_line(self) -> str:
        while True:
            if self._lineno >= len(self._lines):
                raise EndOfTextException
            line = self._lines[self._lineno]
            self._next_line()
            if len(line) > 0:
                return line
            # if no non-empty line, will always break on EndOfLineException

    def _parse_var(self, var_str: str) -> Union[str, IqoalaVector]:
        if var_str.startswith("vec<"):
            vec_values_str = var_str[4:-1]
            if len(vec_values_str) == 0:
                vec_values = []
            else:
                vec_values = [x.strip() for x in vec_values_str.split(";")]
            return IqoalaVector(vec_values)
        else:
            return var_str

    def _parse_lhr(self) -> ClassicalIqoalaOp:
        line = self._read_line()

        attr: Optional[IqoalaValue]

        assign_parts = [x.strip() for x in line.split("=")]
        assert len(assign_parts) <= 2
        if len(assign_parts) == 1:
            value = assign_parts[0]
            result = None
        elif len(assign_parts) == 2:
            value = assign_parts[1]
            result = self._parse_var(assign_parts[0])
        value_parts = [x.strip() for x in value.split(":")]
        assert len(value_parts) <= 2
        if len(value_parts) == 2:
            value = value_parts[0]
            attr_str = value_parts[1]
            try:
                attr = int(attr_str)
            except ValueError:
                attr = attr_str
        else:
            value = value_parts[0]
            attr = None

        op_parts = [x.strip() for x in value.split("(")]
        assert len(op_parts) == 2
        op = op_parts[0]
        arguments = op_parts[1].rstrip(")")
        if len(arguments) == 0:
            raw_args = []
        else:
            raw_args = [x.strip() for x in arguments.split(",")]

        args = [self._parse_var(arg) for arg in raw_args]

        # print(f"result = {result}, op = {op}, args = {args}, attr = {attr}")

        lhr_op = LHR_OP_NAMES[op].from_generic_args(result, args, attr)
        return lhr_op

    def parse(self) -> List[ClassicalIqoalaOp]:
        instructions: List[ClassicalIqoalaOp] = []

        try:
            while True:
                instr = self._parse_lhr()
                instructions.append(instr)
        except AssertionError:
            raise IqoalaParseError
        except EndOfTextException:
            pass

        return instructions


class IQoalaSubroutineParser:
    def __init__(self, text: str) -> None:
        self._text = text
        lines = [line.strip() for line in text.split("\n")]
        self._lines = [line for line in lines if len(line) > 0]
        self._lineno: int = 0

    def _next_line(self) -> None:
        self._lineno += 1

    def _read_line(self) -> str:
        while True:
            if self._lineno >= len(self._lines):
                raise EndOfTextException
            line = self._lines[self._lineno]
            self._next_line()
            if len(line) > 0:
                return line
            # if no non-empty line, will always break on EndOfLineException

    def _parse_subrt_meta_line(self, key: str, line: str) -> List[str]:
        split = line.split(":")
        assert len(split) >= 1
        assert split[0] == key
        if len(split) == 1:
            return []
        assert len(split) == 2
        if len(split[1]) == 0:
            return []
        values = split[1].split(",")
        return [v.strip() for v in values]

    def _parse_nqasm_return_mapping(
        self, value_str: str
    ) -> Dict[str, IqoalaSharedMemLoc]:
        result_dict = {}
        for v in value_str:
            key_value = [x.strip() for x in v.split("->")]
            assert len(key_value) == 2
            result_dict[key_value[1]] = IqoalaSharedMemLoc(key_value[0])
        return result_dict

    def _parse_subroutine(self) -> IqoalaSubroutine:
        return_map: Dict[str, IqoalaSharedMemLoc] = {}
        name_line = self._read_line()
        assert name_line.startswith("SUBROUTINE ")
        name = name_line[len("SUBROUTINE") + 1 :]
        params_line = self._parse_subrt_meta_line("params", self._read_line())
        # TODO: use params line?
        return_map_line = self._parse_subrt_meta_line("returns", self._read_line())
        return_map = self._parse_nqasm_return_mapping(return_map_line)
        request_line = self._parse_subrt_meta_line("request", self._read_line())
        assert len(request_line) in [0, 1]
        request_name = None if len(request_line) == 0 else request_line[0]

        start_line = self._read_line()
        assert start_line == "NETQASM_START"
        subrt_lines = []
        while True:
            line = self._read_line()
            if line == "NETQASM_END":
                break
            subrt_lines.append(line)
        subrt_text = "\n".join(subrt_lines)
        try:
            subrt = parse_text_subroutine(subrt_text)
        except KeyError:
            subrt = parse_text_subroutine(subrt_text, flavour=NVFlavour())

        # Check that all templates are declared as params to the subroutine
        if any(arg not in params_line for arg in subrt.arguments):
            raise IqoalaParseError
        return IqoalaSubroutine(name, subrt, return_map, request_name)

    def parse(self) -> Dict[str, IqoalaSubroutine]:
        subroutines: Dict[str, IqoalaSubroutine] = {}
        try:
            while True:
                subrt = self._parse_subroutine()
                subroutines[subrt.name] = subrt
        except EndOfTextException:
            return subroutines


class IQoalaRequestParser:
    def __init__(self, text: str) -> None:
        self._text = text
        lines = [line.strip() for line in text.split("\n")]
        self._lines = [line for line in lines if len(line) > 0]
        self._lineno: int = 0

    def _next_line(self) -> None:
        self._lineno += 1

    def _read_line(self) -> str:
        while True:
            if self._lineno >= len(self._lines):
                raise EndOfTextException
            line = self._lines[self._lineno]
            self._next_line()
            if len(line) > 0:
                return line
            # if no non-empty line, will always break on EndOfLineException

    def _parse_request_line(self, key: str, line: str) -> List[str]:
        split = line.split(":")
        assert len(split) >= 1
        assert split[0] == key
        if len(split) == 1:
            return []
        assert len(split) == 2
        if len(split[1]) == 0:
            return []
        values = split[1].split(",")
        return [v.strip() for v in values]

    def _parse_single_int_value(
        self, key: str, line: str, allow_template: bool = False
    ) -> Union[int, Template]:
        strings = self._parse_request_line(key, line)
        if len(strings) != 1:
            raise IqoalaParseError
        value = strings[0]
        if allow_template:
            if value.startswith("{") and value.endswith("}"):
                value = value.strip("{}").strip()
                return Template(value)
        return int(value)

    def _parse_int_list_value(self, key: str, line: str) -> int:
        strings = self._parse_request_line(key, line)
        return [int(s) for s in strings]

    def _parse_single_float_value(self, key: str, line: str) -> int:
        strings = self._parse_request_line(key, line)
        if len(strings) != 1:
            raise IqoalaParseError
        return float(strings[0])

    def _parse_epr_create_role_value(self, key: str, line: str) -> int:
        strings = self._parse_request_line(key, line)
        if len(strings) != 1:
            raise IqoalaParseError
        try:
            return EprCreateRole[strings[0].upper()]
        except KeyError:
            raise IqoalaParseError

    def _parse_epr_create_type_value(self, key: str, line: str) -> int:
        strings = self._parse_request_line(key, line)
        if len(strings) != 1:
            raise IqoalaParseError
        try:
            return EprCreateType[strings[0].upper()]
        except KeyError:
            raise IqoalaParseError

    def _parse_request(self) -> IqoalaRequest:
        name_line = self._read_line()
        if not name_line.startswith("REQUEST "):
            raise IqoalaParseError
        name = name_line[len("REQUEST") + 1 :]

        role = self._parse_epr_create_role_value("role", self._read_line())
        remote_id = self._parse_single_int_value(
            "remote_id", self._read_line(), allow_template=True
        )
        epr_socket_id = self._parse_single_int_value("epr_socket_id", self._read_line())
        typ = self._parse_epr_create_type_value("typ", self._read_line())
        num_pairs = self._parse_single_int_value("num_pairs", self._read_line())
        fidelity = self._parse_single_float_value("fidelity", self._read_line())
        virt_qubit_ids = self._parse_int_list_value("virt_qubit_ids", self._read_line())
        result_array_addr = self._parse_single_int_value(
            "result_array_addr", self._read_line()
        )

        if role == EprCreateRole.CREATE:
            request = NetstackCreateRequest(
                remote_id=remote_id,
                epr_socket_id=epr_socket_id,
                typ=typ,
                num_pairs=num_pairs,
                fidelity=fidelity,
                virt_qubit_ids=virt_qubit_ids,
                result_array_addr=result_array_addr,
            )
        else:
            assert role == EprCreateRole.RECEIVE
            request = NetstackReceiveRequest(
                remote_id=remote_id,
                epr_socket_id=epr_socket_id,
                typ=typ,
                num_pairs=num_pairs,
                fidelity=fidelity,
                virt_qubit_ids=virt_qubit_ids,
                result_array_addr=result_array_addr,
            )

        return IqoalaRequest(name=name, role=role, request=request)

    def parse(self) -> Dict[str, IqoalaRequest]:
        requests: Dict[str, IqoalaRequest] = {}
        try:
            while True:
                request = self._parse_request()
                requests[request.name] = request
        except EndOfTextException:
            return requests


class IqoalaParser:
    def __init__(
        self,
        text: Optional[str] = None,
        meta_text: Optional[str] = None,
        instr_text: Optional[str] = None,
        subrt_text: Optional[str] = None,
        req_text: Optional[str] = None,
    ) -> None:
        if text is not None:
            meta_text, instr_text, subrt_text, req_text = self._split_text(text)
        else:
            assert meta_text is not None
            assert instr_text is not None
            assert subrt_text is not None
            assert req_text is not None
        self._meta_text = meta_text
        self._instr_text = instr_text
        self._subrt_text = subrt_text
        self._req_text = req_text
        self._meta_parser = IqoalaMetaParser(meta_text)
        self._instr_parser = IqoalaInstrParser(instr_text)
        self._subrt_parser = IQoalaSubroutineParser(subrt_text)
        self._req_parser = IQoalaRequestParser(req_text)

    def _split_text(self, text: str) -> Tuple[str, str, str, str]:
        lines = [line.strip() for line in text.split("\n")]
        meta_end_line: int
        first_subrt_line: int
        first_req_line: Optional[int] = None
        for i, line in enumerate(lines):
            if "META_END" in line:
                meta_end_line = i
                break
        for i, line in enumerate(lines):
            if "SUBROUTINE" in line:
                first_subrt_line = i
                break
        for i, line in enumerate(lines):
            if "REQUEST" in line:
                first_req_line = i
                break

        meta_text = "\n".join(lines[0 : meta_end_line + 1])
        instr_text = "\n".join(lines[meta_end_line + 1 : first_subrt_line])
        if first_req_line is None:
            subrt_text = "\n".join(lines[first_subrt_line:])
            req_text = ""
        else:
            subrt_text = "\n".join(lines[first_subrt_line:first_req_line])
            req_text = "\n".join(lines[first_req_line:])

        return meta_text, instr_text, subrt_text, req_text

    def parse(self) -> IqoalaProgram:
        instructions = self._instr_parser.parse()
        subroutines = self._subrt_parser.parse()
        requests = self._req_parser.parse()
        meta = self._meta_parser.parse()

        # Check that all references to subroutines (in RunSubroutineOp instructions)
        # are valid.
        for instr in instructions:
            if isinstance(instr, RunSubroutineOp):
                subrt_name = instr.subroutine
                if subrt_name not in subroutines:
                    raise IqoalaParseError
        return IqoalaProgram(instructions, subroutines, meta, requests)



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Dict, List

import netsquid as ns

from qoala.lang.iqoala import IqoalaParser, IqoalaProgram
from qoala.runtime.config import (
    DepolariseLinkConfig,
    GenericQDeviceConfig,
    LinkConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
)
from qoala.runtime.environment import GlobalEnvironment, GlobalNodeInfo
from qoala.runtime.program import BatchInfo, BatchResult, ProgramBatch, ProgramInput
from qoala.runtime.schedule import (
    NaiveSolver,
    NoTimeSolver,
    ProgramTaskList,
    TaskBuilder,
)
from qoala.sim.build import build_network
from qoala.sim.network import ProcNodeNetwork

INSTR_LATENCY = 1e5
CC_LATENCY = 1e6
QC_EXPECTATION = 30e6


def create_global_env(
    num_clients: int, global_schedule: List[int], timeslot_len: int
) -> GlobalEnvironment:

    env = GlobalEnvironment()
    env.add_node(0, GlobalNodeInfo("server", 0))
    for i in range(1, num_clients + 1):
        env.add_node(i, GlobalNodeInfo(f"client_{i}", i))

    env.set_global_schedule(global_schedule)
    env.set_timeslot_len(timeslot_len)
    return env


def get_client_config(id: int) -> ProcNodeConfig:
    # client only needs 1 qubit
    return ProcNodeConfig(
        name=f"client_{id}",
        node_id=id,
        qdevice_typ="generic",
        qdevice_cfg=GenericQDeviceConfig.perfect_config(1),
        instr_latency=INSTR_LATENCY,
        receive_latency=CC_LATENCY,
    )


def get_server_config(id: int, num_qubits: int) -> ProcNodeConfig:
    config_file = relative_to_cwd("node_config.yaml")
    qdevice_cfg = GenericQDeviceConfig.from_file(config_file)
    qdevice_cfg.num_qubits = num_qubits
    return ProcNodeConfig(
        name="server",
        node_id=id,
        qdevice_typ="generic",
        # qdevice_cfg=GenericQDeviceConfig.perfect_config(num_qubits),
        qdevice_cfg=qdevice_cfg,
        instr_latency=INSTR_LATENCY,
        receive_latency=CC_LATENCY,
    )


def create_network(
    server_cfg: ProcNodeConfig,
    client_configs: List[ProcNodeConfig],
    num_clients: int,
    global_schedule: List[int],
    timeslot_len: int,
) -> ProcNodeNetwork:
    assert len(client_configs) == num_clients

    global_env = create_global_env(num_clients, global_schedule, timeslot_len)

    link_cfg_file = relative_to_cwd("link_config.yaml")
    depolarise_config = DepolariseLinkConfig.from_file(link_cfg_file)
    link_cfgs = [
        LinkConfig(
            node1="server", node2=cfg.name, typ="depolarise", cfg=depolarise_config
        )
        for cfg in client_configs
    ]

    node_cfgs = [server_cfg] + client_configs

    network_cfg = ProcNodeNetworkConfig(nodes=node_cfgs, links=link_cfgs)
    return build_network(network_cfg, global_env)


@dataclass
class TaskDurations:
    instr_latency: int
    cc_latency: int
    single_gate: int
    two_gate: int
    meas: int


def create_server_tasks(
    server_program: IqoalaProgram, task_durations: TaskDurations
) -> ProgramTaskList:
    tasks = []

    cl_dur = 2 * task_durations.instr_latency
    cc_dur = 2 * task_durations.cc_latency
    qc_dur = QC_EXPECTATION

    set_dur = task_durations.instr_latency
    rot_dur = task_durations.single_gate
    h_dur = task_durations.single_gate
    meas_dur = task_durations.meas
    free_dur = task_durations.instr_latency
    cphase_dur = task_durations.two_gate

    # csocket = assign_cval() : 0
    task0 = TaskBuilder.CL(cl_dur, 0)
    # run_subroutine(vec<client_id>) : create_epr_0
    task1 = TaskBuilder.CL(cl_dur, 1)
    task2 = TaskBuilder.QC(qc_dur, "create_epr_0")
    # run_subroutine(vec<client_id>) : create_epr_1
    task3 = TaskBuilder.CL(cl_dur, 2)
    task4 = TaskBuilder.QC(qc_dur, "create_epr_1")

    # run_subroutine(vec<client_id>) : local_cphase
    task5 = TaskBuilder.CL(cl_dur, 3)
    task6 = TaskBuilder.QL(set_dur, "local_cphase", 0)
    task7 = TaskBuilder.QL(set_dur, "local_cphase", 1)
    task8 = TaskBuilder.QL(cphase_dur, "local_cphase", 2)

    # delta1 = recv_cmsg(client_id)
    task9 = TaskBuilder.CC(cc_dur, 4)

    # vec<m1> = run_subroutine(vec<delta1>) : meas_qubit_1
    task10 = TaskBuilder.CL(cl_dur, 5)
    task11 = TaskBuilder.QL(set_dur, "meas_qubit_1", 0)
    task12 = TaskBuilder.QL(rot_dur, "meas_qubit_1", 1)
    task13 = TaskBuilder.QL(h_dur, "meas_qubit_1", 2)
    task14 = TaskBuilder.QL(meas_dur, "meas_qubit_1", 3)
    task15 = TaskBuilder.QL(free_dur, "meas_qubit_1", 4)

    # send_cmsg(csocket, m1)
    task16 = TaskBuilder.CC(cc_dur, 6)

    # delta2 = recv_cmsg(csocket)
    task17 = TaskBuilder.CC(cc_dur, 7)

    # vec<m2> = run_subroutine(vec<delta2>) : meas_qubit_0
    task18 = TaskBuilder.CL(cl_dur, 8)
    task19 = TaskBuilder.QL(set_dur, "meas_qubit_0", 0)
    task20 = TaskBuilder.QL(rot_dur, "meas_qubit_0", 1)
    task21 = TaskBuilder.QL(h_dur, "meas_qubit_0", 2)
    task22 = TaskBuilder.QL(meas_dur, "meas_qubit_0", 3)
    task23 = TaskBuilder.QL(free_dur, "meas_qubit_0", 4)

    # send_cmsg(csocket, m2)
    task24 = TaskBuilder.CC(cc_dur, 9)

    tasks.append(TaskBuilder.task_group([task0, task1]))
    tasks.append(task2)
    tasks.append(task3)
    tasks.append(task4)
    tasks.append(task5)
    tasks.append(TaskBuilder.task_group([task6, task7, task8]))
    tasks.append(task9)
    tasks.append(task10)
    tasks.append(TaskBuilder.task_group([task11, task12, task13, task14, task15]))
    tasks.append(task16)
    tasks.append(task17)
    tasks.append(task18)
    tasks.append(TaskBuilder.task_group([task19, task20, task21, task22, task23]))
    tasks.append(task24)

    return ProgramTaskList(server_program, {i: task for i, task in enumerate(tasks)})


def create_client_tasks(
    client_program: IqoalaProgram, task_durations: TaskDurations
) -> ProgramTaskList:
    tasks = []

    cl_dur = task_durations.instr_latency
    cc_dur = task_durations.cc_latency
    qc_dur = QC_EXPECTATION

    set_dur = task_durations.instr_latency
    rot_dur = task_durations.single_gate
    h_dur = task_durations.single_gate
    meas_dur = task_durations.meas
    free_dur = task_durations.instr_latency

    class Counter:
        def __init__(self):
            self.index = 0

        def next(self):
            index = self.index
            self.index += 1
            return index

    c = Counter()

    # csocket = assign_cval() : 0
    tasks.append(TaskBuilder.CL(cl_dur, c.next()))
    # const_1 = assign_cval() : 1
    tasks.append(TaskBuilder.CL(cl_dur, c.next()))

    # compute epr0_rot_y etc
    for _ in range(10):
        tasks.append(TaskBuilder.CL(cl_dur, c.next()))

    # run_subroutine(vec<>) : create_epr_0
    tasks.append(TaskBuilder.CL(cl_dur, c.next()))
    # vec<p2> = run_subroutine(vec<theta2>) : post_epr_0
    tasks.append(TaskBuilder.QC(qc_dur, "create_epr_0"))

    tasks.append(TaskBuilder.CL(cl_dur, c.next()))
    tasks.append(TaskBuilder.QL(set_dur, "post_epr_0", 0))
    tasks.append(TaskBuilder.QL(rot_dur, "post_epr_0", 1))
    tasks.append(TaskBuilder.QL(h_dur, "post_epr_0", 2))
    tasks.append(TaskBuilder.QL(meas_dur, "post_epr_0", 3))
    tasks.append(TaskBuilder.QL(free_dur, "post_epr_0", 4))
    tasks.append(TaskBuilder.QL(free_dur, "post_epr_0", 5))

    tasks.append(TaskBuilder.CL(cl_dur, c.next()))
    tasks.append(TaskBuilder.QC(qc_dur, "create_epr_1"))
    tasks.append(TaskBuilder.CL(cl_dur, c.next()))
    tasks.append(TaskBuilder.QL(set_dur, "post_epr_1", 0))
    tasks.append(TaskBuilder.QL(rot_dur, "post_epr_1", 1))
    tasks.append(TaskBuilder.QL(h_dur, "post_epr_1", 2))
    tasks.append(TaskBuilder.QL(meas_dur, "post_epr_1", 3))
    tasks.append(TaskBuilder.QL(free_dur, "post_epr_1", 4))
    tasks.append(TaskBuilder.QL(free_dur, "post_epr_1", 5))

    # x = mult_const(p1) : 16
    # minus_theta1 = mult_const(theta1) : -1
    # delta1 = add_cval_c(minus_theta1, x)
    # delta1 = add_cval_c(delta1, alpha)
    for _ in range(4):
        tasks.append(TaskBuilder.CL(cl_dur, c.next()))

    # minus_dummy0 = mult_const(dummy0) : -1
    # should_correct_0 = add_cval_c(const_1, minus_dummy0)
    # delta1_correction = bcond_mult_const(alpha, should_correct_0) : 0
    # delta1_correction = mult_const(delta1_correction) : -1
    # delta1 = add_cval_c(delta1, delta1_correction)
    for _ in range(5):
        tasks.append(TaskBuilder.CL(cl_dur, c.next()))

    # send_cmsg(csocket, delta1)
    # m1 = recv_cmsg(csocket)
    tasks.append(TaskBuilder.CC(cl_dur, c.next()))
    tasks.append(TaskBuilder.CC(cc_dur, c.next()))

    # y = mult_const(p2) : 16
    # minus_theta2 = mult_const(theta2) : -1
    # beta = bcond_mult_const(beta, m1) : -1
    # delta2 = add_cval_c(beta, minus_theta2)
    # delta2 = add_cval_c(delta2, y)
    for _ in range(5):
        tasks.append(TaskBuilder.CL(cl_dur, c.next()))

    # minus_dummy1 = mult_const(dummy1) : -1
    # should_correct_1 = add_cval_c(const_1, minus_dummy1)
    # delta2_correction = bcond_mult_const(beta, should_correct_1) : 0
    # delta2_correction = mult_const(delta2_correction) : -1
    # delta2 = add_cval_c(delta2, delta2_correction)
    for _ in range(5):
        tasks.append(TaskBuilder.CL(cl_dur, c.next()))

    # send_cmsg(csocket, delta2)
    # m2 = recv_cmsg(csocket)
    tasks.append(TaskBuilder.CC(cl_dur, c.next()))
    tasks.append(TaskBuilder.CC(cl_dur, c.next()))

    # return results
    for _ in range(4):
        tasks.append(TaskBuilder.CL(cl_dur, c.next()))

    return ProgramTaskList(client_program, {i: task for i, task in enumerate(tasks)})


@dataclass
class BqcResult:
    client_batches: List[Dict[int, ProgramBatch]]
    client_results: List[Dict[int, BatchResult]]


def create_durations() -> TaskDurations:
    config_file = relative_to_cwd("node_config.yaml")
    qdevice_cfg = GenericQDeviceConfig.from_file(config_file)

    return TaskDurations(
        instr_latency=INSTR_LATENCY,
        cc_latency=CC_LATENCY,
        single_gate=qdevice_cfg.single_qubit_gate_time,
        two_gate=qdevice_cfg.two_qubit_gate_time,
        meas=qdevice_cfg.measure_time,
    )


def relative_to_cwd(file: str) -> str:
    return os.path.join(os.path.dirname(__file__), file)


def load_server_program(remote_name: str) -> IqoalaProgram:
    path = relative_to_cwd("bqc_server.iqoala")
    with open(path) as file:
        server_text = file.read()
    program = IqoalaParser(server_text).parse()

    # Replace "client" by e.g. "client_1"
    program.meta.csockets[0] = remote_name
    program.meta.epr_sockets[0] = remote_name

    return program


def load_client_program() -> IqoalaProgram:
    path = relative_to_cwd("bqc_client.iqoala")
    with open(path) as file:
        client_text = file.read()
    return IqoalaParser(client_text).parse()


def create_server_batch(
    client_id: int,
    inputs: List[ProgramInput],
    num_iterations: int,
    deadline: int,
    num_qubits: int,
) -> BatchInfo:
    durations = create_durations()
    server_program = load_server_program(remote_name=f"client_{client_id}")
    server_tasks = create_server_tasks(server_program, durations)
    return BatchInfo(
        program=server_program,
        inputs=inputs,
        num_iterations=num_iterations,
        deadline=deadline,
        tasks=server_tasks,
        num_qubits=num_qubits,
    )


def create_client_batch(
    inputs: List[ProgramInput], num_iterations: int, deadline: int
) -> BatchInfo:
    durations = create_durations()
    client_program = load_client_program()
    client_tasks = create_client_tasks(client_program, durations)
    return BatchInfo(
        program=client_program,
        inputs=inputs,
        num_iterations=num_iterations,
        deadline=deadline,
        tasks=client_tasks,
        num_qubits=1,
    )


def run_bqc(
    alpha,
    beta,
    theta1,
    theta2,
    dummy0,
    dummy1,
    num_iterations: List[int],
    deadlines: List[int],
    num_clients: int,
    global_schedule: List[int],
    timeslot_len: int,
):
    # server needs to have 2 qubits per client
    server_num_qubits = num_clients * 2
    server_config = get_server_config(id=0, num_qubits=server_num_qubits)
    client_configs = [get_client_config(i) for i in range(1, num_clients + 1)]

    network = create_network(
        server_config, client_configs, num_clients, global_schedule, timeslot_len
    )

    server_procnode = network.nodes["server"]

    for client_id in range(1, num_clients + 1):
        # index in num_iterations and deadlines list
        index = client_id - 1

        server_inputs = [
            ProgramInput({"client_id": client_id}) for _ in range(num_iterations[index])
        ]

        server_batch_info = create_server_batch(
            client_id=client_id,
            inputs=server_inputs,
            num_iterations=num_iterations[index],
            deadline=deadlines[index],
            num_qubits=server_num_qubits,
        )

        server_procnode.submit_batch(server_batch_info)
    server_procnode.initialize_processes()
    server_procnode.initialize_schedule(NaiveSolver)

    for client_id in range(1, num_clients + 1):
        # index in num_iterations and deadlines list
        index = client_id - 1

        client_inputs = [
            ProgramInput(
                {
                    "server_id": 0,
                    "alpha": alpha,
                    "beta": beta,
                    "theta1": theta1,
                    "theta2": theta2,
                    "dummy0": dummy0,
                    "dummy1": dummy1,
                }
            )
            for _ in range(num_iterations[index])
        ]

        client_batch_info = create_client_batch(
            client_inputs, num_iterations[index], deadlines[index]
        )

        client_procnode = network.nodes[f"client_{client_id}"]
        client_procnode.submit_batch(client_batch_info)
        client_procnode.initialize_processes()
        client_procnode.initialize_schedule(NoTimeSolver)

    network.start_all_nodes()
    start_time = ns.sim_time()
    ns.sim_run()
    end_time = ns.sim_time()
    makespan = end_time - start_time

    client_procnodes = [network.nodes[f"client_{i}"] for i in range(1, num_clients + 1)]
    client_batches = [node.get_batches() for node in client_procnodes]

    client_results: List[Dict[int, BatchResult]]
    client_results = [node.scheduler.get_batch_results() for node in client_procnodes]

    return BqcResult(client_batches, client_results), makespan


def check(
    alpha,
    beta,
    theta1,
    theta2,
    dummy0,
    dummy1,
    expected,
    num_iterations,
    deadlines,
    num_clients,
    global_schedule: List[int],
    timeslot_len: int,
):
    ns.sim_reset()
    bqc_result, makespan = run_bqc(
        alpha=alpha,
        beta=beta,
        theta1=theta1,
        theta2=theta2,
        dummy0=dummy0,
        dummy1=dummy1,
        num_iterations=num_iterations,
        deadlines=deadlines,
        num_clients=num_clients,
        global_schedule=global_schedule,
        timeslot_len=timeslot_len,
    )

    batch_success_probabilities: List[float] = []

    for i in range(num_clients):
        assert len(bqc_result.client_results[i]) == 1
        batch_result = bqc_result.client_results[i][0]
        assert len(bqc_result.client_batches[i]) == 1
        program_batch = bqc_result.client_batches[i][0]
        batch_iterations = program_batch.info.num_iterations

        p1s = [result.values["p1"] for result in batch_result.results]
        p2s = [result.values["p2"] for result in batch_result.results]
        m1s = [result.values["m1"] for result in batch_result.results]
        m2s = [result.values["m2"] for result in batch_result.results]

        if dummy0 == 0:
            num_fails = len([(p, m) for (p, m) in zip(p1s, m2s) if p != m])
        else:  # dummy0 = 1
            num_fails = len([(p, m) for (p, m) in zip(p2s, m1s) if p != m])

        frac_fail = round(num_fails / batch_iterations, 2)
        batch_success_probabilities.append(1 - frac_fail)

    return batch_success_probabilities, makespan


def compute_succ_prob(
    num_clients: int,
    num_iterations: List[int],
    deadlines: List[int],
    global_schedule: List[int],
    timeslot_len: int,
):
    ns.set_qstate_formalism(ns.qubits.qformalism.QFormalism.DM)

    return check(
        alpha=8,
        beta=24,
        theta1=2,
        theta2=22,
        dummy0=0,
        dummy1=1,
        expected=1,
        num_iterations=num_iterations,
        deadlines=deadlines,
        num_clients=num_clients,
        global_schedule=global_schedule,
        timeslot_len=timeslot_len,
    )


def test_bqc():
    num_clients = 10
    succ_probs, makespan = compute_succ_prob(
        num_clients=num_clients,
        num_iterations=[30] * num_clients,
        deadlines=[1e8] * num_clients,
        global_schedule=[i for i in range(num_clients)],
        timeslot_len=50e6,
    )
    print(f"success probabilities: {succ_probs}")
    print(f"makespan: {makespan}")


if __name__ == "__main__":
    test_bqc()



from __future__ import annotations

from typing import Generator

from pydynaa import EventExpression
from qoala.sim.common import ComponentProtocol, PortListener
from qoala.sim.events import EVENT_WAIT
from qoala.sim.memmgr import MemoryManager
from qoala.sim.message import Message
from qoala.sim.qdevice import QDevice
from qoala.sim.qnoscomp import QnosComponent
from qoala.sim.signals import (
    SIGNAL_HOST_HAND_MSG,
    SIGNAL_MEMORY_FREED,
    SIGNAL_NSTK_PROC_MSG,
)


class QnosInterface(ComponentProtocol):
    """NetSquid protocol representing a QNodeOS processor."""

    def __init__(
        self, comp: QnosComponent, qdevice: QDevice, memmgr: MemoryManager
    ) -> None:
        """Processor protocol constructor. Typically created indirectly through
        constructing a `Qnos` instance.

        :param comp: NetSquid component representing the processor
        :param qnos: `Qnos` protocol that owns this protocol
        """
        super().__init__(name=f"{comp.name}_protocol", comp=comp)
        self._comp = comp
        self._qdevice = qdevice
        self._memmgr = memmgr

        self.add_listener(
            "host",
            PortListener(self._comp.ports["host_in"], SIGNAL_HOST_HAND_MSG),
        )
        self.add_listener(
            "netstack",
            PortListener(self._comp.ports["nstk_in"], SIGNAL_NSTK_PROC_MSG),
        )

        self.add_signal(SIGNAL_MEMORY_FREED)

    def send_host_msg(self, msg: Message) -> None:
        self._comp.host_out_port.tx_output(msg)

    def receive_host_msg(self) -> Generator[EventExpression, None, Message]:
        return (yield from self._receive_msg("host", SIGNAL_HOST_HAND_MSG))

    def send_netstack_msg(self, msg: Message) -> None:
        self._comp.netstack_out_port.tx_output(msg)

    def receive_netstack_msg(self) -> Generator[EventExpression, None, Message]:
        return (yield from self._receive_msg("netstack", SIGNAL_NSTK_PROC_MSG))

    def flush_netstack_msgs(self) -> None:
        self._listeners["netstack"].buffer.clear()

    def signal_memory_freed(self) -> None:
        self._comp.netstack_mem_out_port.tx_output(Message(content=None))

    @property
    def qdevice(self) -> QDevice:
        return self._qdevice

    @property
    def memmgr(self) -> MemoryManager:
        return self._memmgr

    def wait(self, delta_time: int) -> Generator[EventExpression, None, None]:
        self._schedule_after(delta_time, EVENT_WAIT)
        event_expr = EventExpression(source=self, event_type=EVENT_WAIT)
        yield event_expr



from qoala.sim.memory import CommQubitTrait, MemQubitTrait, Topology, UnitModule


def test_create_unit_module():
    um = UnitModule.from_topology(Topology(comm_ids={0}, mem_ids={0}))
    assert um == UnitModule(
        qubit_ids=[0], qubit_traits={0: [CommQubitTrait, MemQubitTrait]}, gate_traits={}
    )

    um2 = UnitModule.from_topology(Topology(comm_ids={0}, mem_ids={1, 2}))
    assert um2 == UnitModule(
        qubit_ids=[0, 1, 2],
        qubit_traits={0: [CommQubitTrait], 1: [MemQubitTrait], 2: [MemQubitTrait]},
        gate_traits={},
    )


if __name__ == "__main__":
    test_create_unit_module()



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, Generator, List, Optional, Tuple

import pytest
from netqasm.lang.parsing import parse_text_subroutine

from pydynaa import EventExpression
from qoala.lang.iqoala import IqoalaProgram, IqoalaSubroutine, ProgramMeta
from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
from qoala.runtime.schedule import ProgramTaskList
from qoala.sim.memmgr import AllocError, MemoryManager
from qoala.sim.memory import ProgramMemory, Topology, UnitModule
from qoala.sim.message import Message
from qoala.sim.process import IqoalaProcess
from qoala.sim.qdevice import PhysicalQuantumMemory, QDevice
from qoala.sim.qnosinterface import QnosInterface
from qoala.sim.qnosprocessor import GenericProcessor, QnosProcessor
from qoala.util.tests import yield_from

MOCK_MESSAGE = Message(content=42)
MOCK_QNOS_RET_REG = "R0"
MOCK_QNOS_RET_VALUE = 7


@dataclass(eq=True, frozen=True)
class InterfaceEvent:
    peer: str
    msg: Message


@dataclass(eq=True, frozen=True)
class FlushEvent:
    pass


@dataclass(eq=True, frozen=True)
class SignalEvent:
    pass


class MockQDevice(QDevice):
    def __init__(self, topology: Topology) -> None:
        self._memory = PhysicalQuantumMemory(topology.comm_ids, topology.mem_ids)

    def set_mem_pos_in_use(self, id: int, in_use: bool) -> None:
        pass


@dataclass
class MockNetstackResultInfo:
    pid: int
    array_id: int
    start_idx: int
    end_idx: int


class MockQnosInterface(QnosInterface):
    def __init__(
        self,
        qdevice: QDevice,
        netstack_result_info: Optional[MockNetstackResultInfo] = None,
    ) -> None:
        self.send_events: List[InterfaceEvent] = []
        self.recv_events: List[InterfaceEvent] = []
        self.flush_events: List[FlushEvent] = []
        self.signal_events: List[SignalEvent] = []

        self._qdevice = qdevice
        self._memmgr = MemoryManager("alice", self._qdevice)

        self.netstack_result_info: Optional[
            MockNetstackResultInfo
        ] = netstack_result_info

    def send_peer_msg(self, peer: str, msg: Message) -> None:
        self.send_events.append(InterfaceEvent(peer, msg))

    def receive_peer_msg(self, peer: str) -> Generator[EventExpression, None, Message]:
        self.recv_events.append(InterfaceEvent(peer, MOCK_MESSAGE))
        return MOCK_MESSAGE
        yield  # to make it behave as a generator

    def send_host_msg(self, msg: Message) -> None:
        self.send_events.append(InterfaceEvent("host", msg))

    def receive_host_msg(self) -> Generator[EventExpression, None, Message]:
        self.recv_events.append(InterfaceEvent("host", MOCK_MESSAGE))
        return MOCK_MESSAGE
        yield  # to make it behave as a generator

    def send_netstack_msg(self, msg: Message) -> None:
        self.send_events.append(InterfaceEvent("netstack", msg))

    def receive_netstack_msg(self) -> Generator[EventExpression, None, Message]:
        self.recv_events.append(InterfaceEvent("netstack", MOCK_MESSAGE))
        if self.netstack_result_info is not None:
            mem = self.memmgr._processes[
                self.netstack_result_info.pid
            ].prog_memory.shared_mem
            array_id = self.netstack_result_info.array_id
            start_idx = self.netstack_result_info.start_idx
            end_idx = self.netstack_result_info.end_idx
            for i in range(start_idx, end_idx):
                mem.set_array_value(array_id, i, 42)
        return MOCK_MESSAGE
        yield  # to make it behave as a generator

    def flush_netstack_msgs(self) -> None:
        self.flush_events.append(FlushEvent())

    def signal_memory_freed(self) -> None:
        self.signal_events.append(SignalEvent())

    @property
    def name(self) -> str:
        return "mock"


def create_program(
    subroutines: Optional[Dict[str, IqoalaSubroutine]] = None,
    meta: Optional[ProgramMeta] = None,
) -> IqoalaProgram:
    if subroutines is None:
        subroutines = {}
    if meta is None:
        meta = ProgramMeta.empty("prog")
    return IqoalaProgram(instructions=[], subroutines=subroutines, meta=meta)


def create_process(
    pid: int, program: IqoalaProgram, unit_module: UnitModule
) -> IqoalaProcess:
    instance = ProgramInstance(
        pid=pid,
        program=program,
        inputs=ProgramInput({}),
        tasks=ProgramTaskList.empty(program),
    )
    mem = ProgramMemory(pid=pid, unit_module=unit_module)

    process = IqoalaProcess(
        prog_instance=instance,
        prog_memory=mem,
        csockets={},
        epr_sockets=program.meta.epr_sockets,
        subroutines=program.subroutines,
        requests={},
        result=ProgramResult(values={}),
    )
    return process


def create_process_with_subrt(
    pid: int, subrt_text: str, unit_module: UnitModule
) -> IqoalaProcess:
    subrt = parse_text_subroutine(subrt_text)
    iqoala_subrt = IqoalaSubroutine("subrt", subrt, return_map={})
    meta = ProgramMeta.empty("alice")
    meta.epr_sockets = {0: "bob"}
    program = create_program(subroutines={"subrt": iqoala_subrt}, meta=meta)
    return create_process(pid, program, unit_module)


def execute_process(processor: GenericProcessor, process: IqoalaProcess) -> int:
    subroutines = process.prog_instance.program.subroutines
    netqasm_instructions = subroutines["subrt"].subroutine.instructions

    instr_count = 0

    instr_idx = 0
    while instr_idx < len(netqasm_instructions):
        instr_count += 1
        instr_idx = yield_from(processor.assign(process, "subrt", instr_idx))
    return instr_count


def execute_multiple_processes(
    processor: GenericProcessor, processes: List[IqoalaProcess]
) -> None:
    for proc in processes:
        subroutines = proc.prog_instance.program.subroutines
        netqasm_instructions = subroutines["subrt"].subroutine.instructions
        for i in range(len(netqasm_instructions)):
            yield_from(processor.assign(proc, "subrt", i))


def setup_components(
    topology: Topology,
    netstack_result: Optional[MockNetstackResultInfo] = None,
    asynchronous: bool = False,
) -> Tuple[QnosProcessor, UnitModule]:
    qdevice = MockQDevice(topology)
    unit_module = UnitModule.from_topology(topology)
    interface = MockQnosInterface(qdevice, netstack_result)
    processor = QnosProcessor(interface, asynchronous)
    return (processor, unit_module)


def verify_native(subrt_text: str, num_instr: int) -> bool:
    # check that the instructions in the subroutine text do not expand
    # to additional instructions when parsed and compiled
    parsed_subrt = parse_text_subroutine(subrt_text)
    assert len(parsed_subrt.instructions) == num_instr


def test_set_reg():
    processor, unit_module = setup_components(Topology(comm_ids={0}, mem_ids={1}))

    subrt = """
    set R0 17
    """
    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)
    assert process.prog_memory.shared_mem.get_reg_value("R0") == 17


def test_add():
    processor, unit_module = setup_components(Topology(comm_ids={0}, mem_ids={1}))

    subrt = """
    set R0 2
    set R1 5
    add R2 R0 R1
    """
    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)
    assert process.prog_memory.shared_mem.get_reg_value("R2") == 7


def test_alloc_qubit():
    processor, unit_module = setup_components(Topology(comm_ids={0}, mem_ids={1}))

    subrt = """
    set Q0 0
    qalloc Q0
    """
    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)

    assert processor._interface.memmgr.phys_id_for(process.pid, 0) == 0
    assert processor._interface.memmgr.phys_id_for(process.pid, 1) is None


def test_free_qubit():
    processor, unit_module = setup_components(Topology(comm_ids={0}, mem_ids={1}))

    subrt = """
    set Q0 0
    qalloc Q0
    qfree Q0
    """
    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)

    assert processor._interface.memmgr.phys_id_for(process.pid, 0) is None
    assert processor._interface.memmgr.phys_id_for(process.pid, 1) is None


def test_free_non_allocated():
    processor, unit_module = setup_components(Topology(comm_ids={0}, mem_ids={1}))

    subrt = """
    set Q0 0
    qfree Q0
    """
    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)

    with pytest.raises(AllocError):
        execute_process(processor, process)


def test_alloc_multiple():
    processor, unit_module = setup_components(Topology(comm_ids={0}, mem_ids={1}))

    subrt = """
    set Q0 0
    set Q1 1
    qalloc Q0
    qalloc Q1
    """
    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)

    assert processor._interface.memmgr.phys_id_for(process.pid, 0) == 0
    assert processor._interface.memmgr.phys_id_for(process.pid, 1) == 1


def test_alloc_multiprocess():
    processor, unit_module = setup_components(Topology(comm_ids={0}, mem_ids={1}))

    subrt0 = """
    set Q0 0
    qalloc Q0
    """
    subrt1 = """
    set Q1 1
    qalloc Q1
    """
    process0 = create_process_with_subrt(0, subrt0, unit_module)
    process1 = create_process_with_subrt(1, subrt1, unit_module)
    processor._interface.memmgr.add_process(process0)
    processor._interface.memmgr.add_process(process1)
    execute_multiple_processes(processor, [process0, process1])

    assert processor._interface.memmgr.phys_id_for(process0.pid, 0) == 0
    assert processor._interface.memmgr.phys_id_for(process0.pid, 1) is None
    assert processor._interface.memmgr.phys_id_for(process1.pid, 0) is None
    assert processor._interface.memmgr.phys_id_for(process1.pid, 1) == 1

    assert processor._interface.memmgr._physical_mapping[0].pid == process0.pid
    assert processor._interface.memmgr._physical_mapping[1].pid == process1.pid


def test_alloc_multiprocess_same_virt_id():
    processor, unit_module = setup_components(Topology(comm_ids={0, 1}, mem_ids={0, 1}))

    subrt0 = """
    set Q0 0
    qalloc Q0
    """
    subrt1 = """
    set Q0 0
    qalloc Q0
    """

    process0 = create_process_with_subrt(0, subrt0, unit_module)
    process1 = create_process_with_subrt(1, subrt1, unit_module)
    processor._interface.memmgr.add_process(process0)
    processor._interface.memmgr.add_process(process1)
    execute_multiple_processes(processor, [process0, process1])

    assert processor._interface.memmgr.phys_id_for(process0.pid, 0) == 0
    assert processor._interface.memmgr.phys_id_for(process0.pid, 1) is None
    assert processor._interface.memmgr.phys_id_for(process1.pid, 0) == 1
    assert processor._interface.memmgr.phys_id_for(process1.pid, 1) is None

    assert processor._interface.memmgr._physical_mapping[0].pid == process0.pid
    assert processor._interface.memmgr._physical_mapping[1].pid == process1.pid


def test_alloc_multiprocess_same_virt_id_trait_not_available():
    processor, unit_module = setup_components(Topology(comm_ids={0}, mem_ids={0, 1}))

    subrt0 = """
    set Q0 0
    qalloc Q0
    """
    subrt1 = """
    set Q0 0
    qalloc Q0
    """

    process0 = create_process_with_subrt(0, subrt0, unit_module)
    process1 = create_process_with_subrt(1, subrt1, unit_module)
    processor._interface.memmgr.add_process(process0)
    processor._interface.memmgr.add_process(process1)

    with pytest.raises(AllocError):
        execute_multiple_processes(processor, [process0, process1])


def test_no_branch():
    processor, unit_module = setup_components(Topology(comm_ids={0}, mem_ids={0}))

    subrt = """
    set R3 3
    set R0 0
    beq R3 R0 LABEL1
    set R1 1
    add C0 R3 R1
LABEL1:
    """
    verify_native(subrt, 5)

    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)
    instr_count = execute_process(processor, process)

    assert instr_count == 5
    assert process.prog_memory.shared_mem.get_reg_value("C0") == 4


def test_branch():
    processor, unit_module = setup_components(Topology(comm_ids={0}, mem_ids={0}))

    subrt = """
    set R3 3
    set C3 3
    beq R3 C3 LABEL1
    set R1 1
    add C0 R3 R1
LABEL1:
    """
    verify_native(subrt, 5)

    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)
    instr_count = execute_process(processor, process)

    assert instr_count == 3
    assert process.prog_memory.shared_mem.get_reg_value("C0") == 0


def test_array():
    processor, unit_module = setup_components(Topology(comm_ids={0}, mem_ids={0}))

    subrt = """
    set C10 10
    array C10 @0
    set R4 4
    set C8 8
    store C8 @0[R4]
    """
    verify_native(subrt, 5)

    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)
    instr_count = execute_process(processor, process)

    assert instr_count == 5
    array = process.prog_memory.shared_mem.get_array(0)
    assert len(array) == 10
    assert all(
        process.prog_memory.shared_mem.get_array_value(0, i) is None
        for i in range(10)
        if i != 4
    )
    assert process.prog_memory.shared_mem.get_array_value(0, 4) == 8


# TODO: decide on role of create_epr and recv_epr in qnosprocessor

# def test_create_epr():
#     processor, unit_module = setup_components(Topology(comm_ids={0}, mem_ids={0}))

#     qubit_array = 0
#     arg_array = 1
#     result_array = 2

#     qubit_id = 8
#     create_type = 2  # RSP
#     epr_count = 6
#     rot_x_remote2 = 1

#     remote_id = 2
#     epr_sck_id = 4

#     pid = 0
#     # fidelity = 0.75

#     subrt = f"""
#     array 1 @{qubit_array}
#     store {qubit_id} @{qubit_array}[0]
#     array 20 @{arg_array}
#     store {create_type} @{arg_array}[{SER_CREATE_IDX_TYPE}]
#     store {epr_count} @{arg_array}[{SER_CREATE_IDX_NUMBER}]
#     store {rot_x_remote2} @{arg_array}[{SER_CREATE_IDX_ROTATION_X_REMOTE2}]
#     array 10 @{result_array}
#     create_epr({remote_id}, {epr_sck_id}) 0 1 2
#     """
#     process = create_process_with_subrt(pid, subrt, unit_module)
#     processor._interface.memmgr.add_process(process)
#     execute_process(processor, process)

#     expected_request = NetstackCreateRequest(
#         remote_id=remote_id,
#         epr_socket_id=epr_sck_id,
#         typ=EprCreateType.REMOTE_STATE_PREP,
#         num_pairs=epr_count,
#         fidelity=1.0,  # TODO: fix hardcoded only allowed value
#         virt_qubit_ids=[8],
#         result_array_addr=result_array,
#     )

#     assert process.prog_memory.requests[0] == expected_request

#     mem = process.prog_memory.shared_mem
#     assert mem.get_array_value(qubit_array, 0) == qubit_id
#     assert mem.get_array_value(arg_array, SER_CREATE_IDX_TYPE) == create_type
#     assert mem.get_array_value(arg_array, SER_CREATE_IDX_NUMBER) == epr_count
#     assert (
#         mem.get_array_value(arg_array, SER_CREATE_IDX_ROTATION_X_REMOTE2)
#         == rot_x_remote2
#     )


# def test_create_epr_async():
#     processor, unit_module = setup_components(
#         Topology(comm_ids={0}, mem_ids={0}), asynchronous=True
#     )

#     qubit_array = 0
#     arg_array = 1
#     result_array = 2

#     qubit_id = 8
#     create_type = 2  # RSP
#     epr_count = 6
#     rot_x_remote2 = 1

#     remote_id = 2
#     epr_sck_id = 4

#     pid = 0
#     # fidelity = 0.75

#     subrt = f"""
#     array 1 @{qubit_array}
#     store {qubit_id} @{qubit_array}[0]
#     array 20 @{arg_array}
#     store {create_type} @{arg_array}[{SER_CREATE_IDX_TYPE}]
#     store {epr_count} @{arg_array}[{SER_CREATE_IDX_NUMBER}]
#     store {rot_x_remote2} @{arg_array}[{SER_CREATE_IDX_ROTATION_X_REMOTE2}]
#     array 10 @{result_array}
#     create_epr({remote_id}, {epr_sck_id}) 0 1 2
#     """
#     process = create_process_with_subrt(pid, subrt, unit_module)
#     processor._interface.memmgr.add_process(process)
#     execute_process(processor, process)

#     expected_request = NetstackCreateRequest(
#         remote_id=remote_id,
#         epr_socket_id=epr_sck_id,
#         typ=EprCreateType.REMOTE_STATE_PREP,
#         num_pairs=epr_count,
#         fidelity=1.0,  # TODO: fix hardcoded only allowed value
#         virt_qubit_ids=[8],
#         result_array_addr=result_array,
#     )

#     assert processor._interface.send_events[0] == InterfaceEvent(
#         "netstack", Message(expected_request)
#     )

#     mem = process.prog_memory.shared_mem
#     assert mem.get_array_value(qubit_array, 0) == qubit_id
#     assert mem.get_array_value(arg_array, SER_CREATE_IDX_TYPE) == create_type
#     assert mem.get_array_value(arg_array, SER_CREATE_IDX_NUMBER) == epr_count
#     assert (
#         mem.get_array_value(arg_array, SER_CREATE_IDX_ROTATION_X_REMOTE2)
#         == rot_x_remote2
#     )


# def test_recv_epr():
#     processor, unit_module = setup_components(Topology(comm_ids={0}, mem_ids={0}))

#     qubit_array = 0
#     result_array = 2

#     qubit_id = 8

#     remote_id = 2
#     epr_sck_id = 4

#     pid = 0

#     subrt = f"""
#     array 1 @{qubit_array}
#     store {qubit_id} @{qubit_array}[0]
#     array 10 @{result_array}
#     recv_epr({remote_id}, {epr_sck_id}) 0 2
#     """
#     process = create_process_with_subrt(pid, subrt, unit_module)
#     processor._interface.memmgr.add_process(process)
#     execute_process(processor, process)

#     expected_request = NetstackReceiveRequest(
#         remote_id=remote_id,
#         epr_socket_id=epr_sck_id,
#         typ=None,  # TODO: fix recv_epr instruction
#         num_pairs=None,  # TODO: fix recv_epr instruction
#         fidelity=1.0,  # TODO: fix hardcoded only allowed value
#         virt_qubit_ids=[8],
#         result_array_addr=result_array,
#     )

#     assert process.prog_memory.requests[0] == expected_request

#     mem = process.prog_memory.shared_mem
#     assert mem.get_array_value(qubit_array, 0) == qubit_id


def test_wait_all():
    pid = 0
    array_id = 3
    start_idx = 5
    end_idx = 9

    # Let the mock interface write some result to the array such that
    # our "wait_all" instruction will unblock
    netstack_result = MockNetstackResultInfo(
        pid=pid, array_id=array_id, start_idx=start_idx, end_idx=end_idx
    )

    processor, unit_module = setup_components(
        Topology(comm_ids={0}, mem_ids={0}), netstack_result
    )

    subrt = f"""
    array 10 @{array_id}
    wait_all @{array_id}[{start_idx}:{end_idx}]
    """
    process = create_process_with_subrt(pid, subrt, unit_module)
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)

    mem = process.prog_memory.shared_mem
    assert all(
        mem.get_array_value(array_id, i) is not None for i in range(start_idx, end_idx)
    )


if __name__ == "__main__":
    test_set_reg()
    test_add()
    test_alloc_qubit()
    test_free_qubit()
    test_free_non_allocated()
    test_alloc_multiple()
    test_alloc_multiprocess()
    test_alloc_multiprocess_same_virt_id()
    test_alloc_multiprocess_same_virt_id_trait_not_available()
    test_no_branch()
    test_branch()
    test_array()
    test_wait_all()



from __future__ import annotations

from netsquid.protocols import Protocol

from qoala.runtime.environment import LocalEnvironment
from qoala.sim.memmgr import MemoryManager
from qoala.sim.qdevice import PhysicalQuantumMemory, QDevice, QDeviceType
from qoala.sim.qnoscomp import QnosComponent
from qoala.sim.qnosinterface import QnosInterface
from qoala.sim.qnosprocessor import GenericProcessor, NVProcessor, QnosProcessor


class Qnos(Protocol):
    """NetSquid protocol representing a QNodeOS instance."""

    def __init__(
        self,
        comp: QnosComponent,
        local_env: LocalEnvironment,
        memmgr: MemoryManager,
        qdevice: QDevice,
        asynchronous: bool = False,
    ) -> None:
        """Qnos protocol constructor.

        :param comp: NetSquid component representing the QNodeOS instance
        :param qdevice_type: hardware type of the QDevice of this node
        """
        super().__init__(name=f"{comp.name}_protocol")

        # References to objects.
        self._comp = comp
        self._local_env = local_env

        # Owned objects.
        self._interface = QnosInterface(comp, qdevice, memmgr)
        self._processor: QnosProcessor
        self._asynchronous = asynchronous

        self.create_processor(qdevice.typ)

    def create_processor(self, qdevice_type: QDeviceType) -> None:
        if qdevice_type == QDeviceType.GENERIC:
            self._processor = GenericProcessor(self._interface, self._asynchronous)
        elif qdevice_type == QDeviceType.NV:
            self._processor = NVProcessor(self._interface, self._asynchronous)
        else:
            raise ValueError

    @property
    def qdevice(self) -> QDevice:
        return self._interface.qdevice

    @qdevice.setter
    def qdevice(self, qdevice: QDevice) -> None:
        self._interface._qdevice = qdevice
        self.create_processor(qdevice.typ)

    @property
    def processor(self) -> QnosProcessor:
        return self._processor

    @processor.setter
    def processor(self, processor: QnosProcessor) -> None:
        self._processor = processor

    @property
    def physical_memory(self) -> PhysicalQuantumMemory:
        return self._interface.qdevice.memory

    def start(self) -> None:
        super().start()
        self._interface.start()

    def stop(self) -> None:
        self._interface.stop()
        super().stop()
